#include "basic/stl-utils.h"
#include "decoder.h"

extern string clusters_file;
extern bool beam_check, use_beam;
extern int beam_size;
extern int max_seg_len;

const double BIG = 1e200;

////////////////////////////////////////////////////////////
// For the beam search.

struct CandidateNode {
  CandidateNode(int tag, CandidateNode *prev) : tag(tag), prev(prev) { }
  int tag;
  CandidateNode *prev;
};

struct Candidate {
  Candidate() : score(0), tail(NULL), n(0) { }
  double score;
  CandidateNode *tail;
  int n; // number of elements in the list

  void set_context(PtContext &c) {
    c.i = n-1;
    assert(tail && tail->prev && tail->prev->prev);
    c.curr_tag = tail->tag;
    c.prev_tag = tail->prev->tag;
    c.prev2_tag = tail->prev->prev->tag;
  }

  void set_tags(IntVec &tags) const {
    tags.resize(n);
    CandidateNode *node = tail;
    for(int i = n-1; i >= 0; i--) {
      assert(node);
      tags[i] = node->tag;
      node = node->prev;
    }
    assert(!node);
  }

  Candidate append(int tag, hash_set<int> &addrs, vector<CandidateNode *> &nodes) const {
    Candidate c = *this;
    c.tail = new CandidateNode(tag, c.tail);
    c.n++;
 
    // Record the new node.
    CandidateNode *node = c.tail;
    int addr = (int)node;
    if(!contains(addrs, addr)) {
      addrs.insert(addr);
      nodes.push_back(node);
    }

    return c;
  }
};

bool operator<(const Candidate &c1, const Candidate &c2) {
  return c1.score > c2.score;
}

typedef vector<Candidate> CandidateVec;

////////////////////////////////////////////////////////////

double Decoder::find_best_segs(FeatureSet &features, Sequence &seq, SegVec &segs, bool use_averaged, int max_seg_len) {
  Decoder decoder(features);
  if(use_beam) { // Not exactly using beam, but using alternative method
    SegVecVec segss;
    DoubleVec scores;
    decoder.find_best_segs_dp_nbest(seq, segss, scores, use_averaged, max_seg_len, 1);
    segs = segss[0];
    return scores[0];
  }
  return decoder.find_best_segs_dp(seq, segs, use_averaged, max_seg_len);
}

double Decoder::find_best_segs_dp(Sequence &seq, SegVec &segs, bool use_averaged, int max_seg_len) {
  // Dynamic programming (semi-Markov CRFs).
  int N = len(seq);
  int S = D.seg_db.size();

  track("find_best_segs_dp()", "N=" << N << ", S=" << S << ", max_seg_len=" << max_seg_len, false);

  DoubleVecVec best_scores; // i, seg -> best score of a segmentation ending with seg at i 
  IntVecVec best_prev_segs; // i, seg -> previous tag that resulted in the corresponding best_scores
  IntVecVec best_seg_lens;   // i, seg -> length of the segment ending at i
  matrix_resize(best_scores, N, S);
  matrix_resize(best_prev_segs, N, S);
  matrix_resize(best_seg_lens, N, S);

  SegContext context(D, seq, use_averaged);
  foridx(i, N) {
    foridx(s, S) {
      int max_n = max_seg_len;
      max_n <?= i+1;
      if(s == D.no_seg) max_n = 1;

      bool is_boundary_word = context.seq[i].word == D.boundary_token->word;
      bool is_boundary_seg = s == D.boundary_token->seg.a;

      // Boundary word must be tagged as boundary tag, and
      // non-boundary word must not be tagged as a boundary tag.
      if(is_boundary_word != is_boundary_seg) {
        best_scores[i][s] = -BIG;
        best_prev_segs[i][s] = -1;
        best_seg_lens[i][s] = 1;
        continue;
      }
      if(is_boundary_word) {
        assert(is_boundary_seg);
        if(i > 0) {
          int r = vector_index_max(best_scores[i-1]);
          best_scores[i][s] = best_scores[i-1][r];
          best_seg_lens[i][s] = 1;
          best_prev_segs[i][s] = r;
        }
        else {
          best_scores[i][s] = 0;
          best_seg_lens[i][s] = 1;
          best_prev_segs[i][s] = -1;
        }
        continue;
      }

      // Non-boundary word, non-boundary seg...
      context.curr_seg = s;

      double best_score = -BIG;
      int best_prev_seg = -1;
      int best_seg_len = -1;

      for(int n = 1; n <= max_n; n++) {
        context.i = i-n+1;
        context.n = n;
        assert(context.i >= 0);

        if(context.seq[context.i].word == D.boundary_token->word) break;

        foridx(r, S) {
          context.prev_seg = r;

          // Compute score given curr_seg = s, prev_seg = r
          assert(i-n >= 0);
          double score = best_scores[i-n][r] + features.compute_score(context);
          // TMP: not compute features on O segments
          //double score = best_scores[i-n][r];
          //if(s != D.no_seg) score += features.compute_score(context);

          if(score > best_score) {
            best_score = score;
            best_prev_seg = r;
            best_seg_len = n;
          }
        }
      }

      best_scores[i][s] = best_score;
      best_prev_segs[i][s] = best_prev_seg;
      best_seg_lens[i][s] = best_seg_len;
    }
  }

  // Recover best segmentation.
  segs.resize(N);
  vector_fill(segs, Seg());
  
  int best_s = vector_index_max(best_scores[N-1]);
  for(int i = N-1, s = best_s; i >= 0;) {
    int r = best_prev_segs[i][s];
    int n = best_seg_lens[i][s];

    assert(s >= 0 && s < S);
    segs[i-n+1] = Seg(s, n);

    s = r;
    i -= n;
  }

  return best_scores[N-1][best_s];
}

struct SegDPStateList;

struct SegDPState {
  SegDPState(const SegDPStateList *parent) :
    parent(parent), score(-BIG), prev_state(NULL) { }
  SegDPState(const SegDPStateList *parent, double score, SegDPState *prev_state) :
    parent(parent), score(score), prev_state(prev_state) { }

  bool operator>(const SegDPState &state) const { return score > state.score; }

  inline int start_i() const;
  inline int length() const;

  const SegDPStateList *parent;
  double score; // score of segmentation ending with label seg at i 
  SegDPState *prev_state; // for backtracking: previous state
};

struct SegDPStateList {
  SegDPStateList() { }
  
  vector<SegDPState> states;
  int nstates;
  int i; // where this segment ends
  int s; // label of this segment

  void init(int nstates, int i, int s) {
    this->nstates = nstates;
    this->i = i; this->s = s;
    foridx(q, nstates) states.push_back(SegDPState(this));
  }

  void improve(double score, SegDPState *prev_state) {
    SegDPState state(this, score + (prev_state ? prev_state->score : 0), prev_state);
    if(nstates == 2) {
      if(state > states[0])      states[1] = states[0], states[0] = state;
      else if(state > states[1]) states[1] = state;
    }
    else if(nstates == 1) {
      if(state > states[0]) states[0] = state;
    }
    else {
      assert(0);
    }
  }
};

inline int SegDPState::start_i() const { return prev_state ? prev_state->parent->i+1 : 0; }
inline int SegDPState::length() const { return parent->i-start_i()+1; }

void Decoder::find_best_segs_dp_nbest(Sequence &seq, SegVecVec &segss, DoubleVec &scores, bool use_averaged, int max_seg_len, int top_n) {
  // Dynamic programming (semi-Markov CRFs).
  // Return the top 2.
  int N = len(seq);
  int S = D.seg_db.size();

  track("find_best_segs_dp_nbest()", "N=" << N << ", S=" << S <<
        ", max_seg_len=" << max_seg_len << ", top_n = " << top_n, false); 

  assert(top_n <= 2);
  vector< vector<SegDPStateList> > best;
  matrix_resize(best, N, S);

  SegContext context(D, seq, use_averaged);
  foridx(i, N) {
    foridx(s, S) {
      best[i][s].init(top_n, i, s);

      int max_n = max_seg_len <? i+1;
      if(s == D.no_seg) max_n = 1; // "O"

      bool is_boundary_word = context.seq[i].word == D.boundary_token->word;
      bool is_boundary_seg = s == D.boundary_token->seg.a;

      // Boundary word must be tagged as boundary seg, and
      // non-boundary word must not be tagged as a boundary tag.
      if(is_boundary_word != is_boundary_seg) {
        best[i][s].improve(-BIG, NULL);
        continue;
      }
      if(is_boundary_word) {
        assert(is_boundary_seg);
        if(i > 0)
          foridx(r, S) foridx(q, best[i-1][r].nstates)
            best[i][s].improve(0, &best[i-1][r].states[q]);
        else
          best[i][s].improve(0, NULL);
        continue;
      }

      // Non-boundary word, non-boundary seg...
      context.curr_seg = s;

      for(int n = 1; n <= max_n; n++) {
        context.i = i-n+1;
        context.n = n;
        assert(context.i >= 0);

        // Can't extend segment to include boundary words
        if(context.seq[context.i].word == D.boundary_token->word) break;

        foridx(r, S) {
          context.prev_seg = r;

          // Compute score given curr_seg = s, prev_seg = r
          assert(i-n >= 0);
          foridx(q, best[i][s].nstates) {
            double score = features.compute_score(context);
            best[i][s].improve(score, &best[i-n][r].states[q]);
          }
        }
      }
    }
  }

  // Pick the top n segmentations.
  segss.resize(top_n);
  scores.resize(top_n);
  foridx(q, top_n) {
    SegVec &segs = segss[q];
    segs.resize(N);
  
    // Assume last segment is always the boundary segment.
    int best_s = D.boundary_token->seg.a;
    SegDPState *state = &best[N-1][best_s].states[q];
    scores[q] = state->score;
    for(; state; state = state->prev_state) {
      int i = state->start_i(), n = state->length();
      segs[i] = Seg(state->parent->s, n);
    }
  }
}

////////////////////////////////////////////////////////////

// Decode seq.
// Put the optimal tags in tags.
double Decoder::find_best_tags(FeatureSet &features, Sequence &seq, IntVec &tags, bool use_averaged) {
  Decoder decoder(features);

  if(beam_check) {
    IntVec tags1, tags2;
    double s1 = decoder.find_best_tags_beam(seq, tags1, use_averaged);
    double s2 = decoder.find_best_tags_trigram(seq, tags2, use_averaged);
    assert2(s1 <= s2, s1 << ' ' << s2);
    if(!feq(s1, s2)) {
      cout << "NOT OPTIMAL " << s1 << ' ' << s2 << endl;
    }
    //assert2(tags1 == tags2, tags1 << " ||| " << tags2);
    tags = tags1;
    return s1;
  }

  int dp_order = features.dp_order;
  if(use_beam)      return decoder.find_best_tags_beam(seq, tags, use_averaged);
  if(dp_order <= 1) return decoder.find_best_tags_unigram(seq, tags, use_averaged);
  if(dp_order == 2) return decoder.find_best_tags_bigram(seq, tags, use_averaged);
  if(dp_order == 3) return decoder.find_best_tags_trigram(seq, tags, use_averaged);

  assert2(false, "Order not supported: " << dp_order);
  return 0;
}

////////////////////////////////////////////////////////////
// Tag decoding

double Decoder::find_best_tags_beam(Sequence &seq, IntVec &tags, bool use_averaged) {
  IntVecVec tagss;
  DoubleVec scores;
  find_best_tags_beam_nbest(seq, tagss, scores, use_averaged, 1);
  tags = tagss[0];
  return scores[0];
}

// Do a beam-search.
void Decoder::find_best_tags_beam_nbest(Sequence &seq, IntVecVec &tagss, DoubleVec &scores, bool use_averaged, int top_n) {
  int N = len(seq);
  int T = D.tag_db.size();

  track("find_best_tags_beam()", "N=" << N << ", T=" << T << ", beam_size = " << beam_size << ", taking top " << top_n, false); 

  CandidateVec candidates;
  candidates.push_back(Candidate());
  PtContext context(D, seq, use_averaged);

  // Keep track of the nodes created so we can deallocate them later.
  hash_set<int> addrs; // Addresses of nodes
  vector<CandidateNode *> nodes;

  foridx(i, N) {
    priority_queue<Candidate> new_candidates;
    forvec(_, const Candidate &, c, candidates) {
      foridx(t, T) {
        // For each candidate c of length i, extend with tag t.

        bool is_boundary_word = context.seq[i].word == D.boundary_token->word;
        bool is_boundary_tag = t == D.boundary_token->tag;
        if(is_boundary_word != is_boundary_tag) continue; // Bad

        // Create new candidate.
        Candidate new_c = c.append(t, addrs, nodes);

        if(!is_boundary_word) {
          new_c.set_context(context);
          new_c.score += features.compute_score(context);
        }

        // Add the new candidate.
        if(len(new_candidates) >= beam_size) {
          // Keep the number of candidates in check.
          // The top of the priority queue should be the smallest element.
          if(new_c.score > new_candidates.top().score) {
            new_candidates.pop();
            new_candidates.push(new_c);
          }
        }
        else
          new_candidates.push(new_c);
      }
    }

    // Take candidates from heap.
    candidates.clear();
    while(!new_candidates.empty()) {
      candidates.push_back(new_candidates.top());
      new_candidates.pop();
    }
  }

  // Pick the top n.
  assert(beam_size >= top_n);
  tagss.resize(top_n);
  scores.resize(top_n);
  foridx(i, top_n) {
    const Candidate &c = candidates[len(candidates)-i-1];
    c.set_tags(tagss[i]);
    scores[i] = c.score; 
  } 

  // Clean up
  forvec(_, CandidateNode *, node, nodes)
    delete node;
}

double Decoder::find_best_tags_unigram(Sequence &seq, IntVec &tags, bool use_averaged) {
  // We can make each decision in the sequence independently.
  int N = len(seq);
  int T = D.tag_db.size();

  track("find_best_tags_unigram()", "N=" << N << ", T=" << T, false); 

  double total_score = 0;
  PtContext context(D, seq, use_averaged);

  tags.resize(N);
  foridx(i, N) {
    context.i = i;

    DoubleInt best(-BIG, -1);
    foridx(t, T) { // curr_tag
      context.curr_tag = t;

      bool is_boundary_word = context.seq[i].word == D.boundary_token->word;
      bool is_boundary_tag = t == D.boundary_token->tag;

      if(is_boundary_word == is_boundary_tag)
        improve(best, DoubleInt(is_boundary_word ? 0 : features.compute_score(context), t));
    }

    total_score += best.first;
    tags[i] = best.second;
  }

  return total_score;
}

double Decoder::find_best_tags_bigram(Sequence &seq, IntVec &tags, bool use_averaged) {
  // Dynamic programming.
  int N = len(seq);
  int T = D.tag_db.size();

  track("find_best_tags_bigram()", "N=" << N << ", T=" << T, false); 

  DoubleVecVec best_scores; // i, tag -> best score of for a tag sequence ending with tag at i 
  IntVecVec best_prev_tags; // i, tag -> previous tag that resulted in the corresponding best_scores
  matrix_resize(best_scores, N, T);
  matrix_resize(best_prev_tags, N, T);

  PtContext context(D, seq, use_averaged);

  foridx(i, N) {
    context.i = i;

    foridx(t, T) {
      bool is_boundary_word = context.seq[i].word == D.boundary_token->word;
      bool is_boundary_tag = t == D.boundary_token->tag;

      // Boundary word must be tagged as boundary tag, and
      // non-boundary word must not be tagged as a boundary tag.
      if(is_boundary_word != is_boundary_tag) {
        best_scores[i][t] = -BIG;
        best_prev_tags[i][t] = -1;
        continue;
      }
      if(is_boundary_word) {
        assert(is_boundary_tag);
        if(i > 0) {
          int s = vector_index_max(best_scores[i-1]);
          best_scores[i][t] = best_scores[i-1][s];
          best_prev_tags[i][t] = s;
        }
        else {
          best_scores[i][t] = 0;
          best_prev_tags[i][t] = -1;
        }
        continue;
      }

      // Non-boundary word, non-boundary tag...
      context.curr_tag = t;

      double best_score = -BIG;
      int best_prev_tag = -1;
      foridx(s, T) {
        context.prev_tag = s;

        // Compute score given curr_tag = t, prev_tag = s
        assert(i > 0);
        double score = best_scores[i-1][s] + features.compute_score(context);

        if(score > best_score) {
          best_score = score;
          best_prev_tag = s;
        }
      }

      best_scores[i][t] = best_score;
      best_prev_tags[i][t] = best_prev_tag;
    }
  }

  // Recover best tag sequence.
  tags.resize(N);
  int best_t = vector_index_max(best_scores[N-1]);
  for(int i = N-1, t = best_t; i >= 0; t = best_prev_tags[i][t], i--) {
    assert2(t >= 0 && t < T, "At i=" << i << ", t=" << t);
    tags[i] = t;
  }

  return best_scores[N-1][best_t];
}

double Decoder::find_best_tags_trigram(Sequence &seq, IntVec &tags, bool use_averaged) {
  // Dynamic programming.
  int N = len(seq);
  int T = D.tag_db.size();

  track("find_best_tags_trigram()", "N=" << N << ", T=" << T, false); 

  DoubleVecVecVec best_scores; // i, tag0, tag1 -> best score of for a tag sequence ending at i 
  IntVecVecVec best_prev2_tags; // i, tag0, tag1 -> previous tag that resulted in the corresponding best_scores
  matrix_resize(best_scores, N, T, T);
  matrix_resize(best_prev2_tags, N, T, T);

  PtContext context(D, seq, use_averaged);

  foridx(i, N) {
    context.i = i;

    foridx(t0, T) { // curr_tag
      context.curr_tag = t0;

      bool is_boundary_word = context.seq[i].word == D.boundary_token->word;
      bool is_boundary_tag = t0 == D.boundary_token->tag;

      foridx(t1, T) { // prev_tag
        context.prev_tag = t1;

        bool prev_is_boundary_word = i == 0 || context.seq[i-1].word == D.boundary_token->word;
        bool prev_is_boundary_tag = t1 == D.boundary_token->tag;

        // Boundary word must be tagged as boundary tag, and
        // non-boundary word must not be tagged as a boundary tag.
        if(is_boundary_word != is_boundary_tag || prev_is_boundary_word != prev_is_boundary_tag) {
          best_scores[i][t0][t1] = -BIG;
          best_prev2_tags[i][t0][t1] = -1;
          continue;
        }

        if(is_boundary_word) {
          assert(is_boundary_tag);
          if(i > 0) {
            int t2 = vector_index_max(best_scores[i-1][t1]);
            best_scores[i][t0][t1] = best_scores[i-1][t1][t2];
            best_prev2_tags[i][t0][t1] = t2;
          }
          else {
            best_scores[i][t0][t1] = 0;
            best_prev2_tags[i][t0][t1] = -1;
          }
          continue;
        }

        // Actual DP step - find the previous tag.
        double best_score = -BIG;
        int best_prev2_tag = -1;
        foridx(t2, T) {
          context.prev2_tag = t2;

          // Compute score given curr_tag = t0, prev_tag = t1, prev2_tag = t2
          assert(i > 0);
          double score = best_scores[i-1][t1][t2] + features.compute_score(context);

          // Break ties randomly.
          if(score > best_score) { // || (feq(score, best_score) && rand() % 2)) {
            best_score = score;
            best_prev2_tag = t2;
          }
        }

        best_scores[i][t0][t1] = best_score;
        best_prev2_tags[i][t0][t1] = best_prev2_tag;
      }
    }
  }

  // Recover best tag sequence.
  tags.resize(N);
  IntPair best_t = matrix_index_max(best_scores[N-1]);
  for(int i = N-1, t0 = best_t.first, t1 = best_t.second; i >= 1; i--) {
    assert2(t0 >= 0 && t0 < T, t0);
    assert2(t1 >= 0 && t1 < T, t1);
    tags[i] = t0;
    tags[i-1] = t1;

    int t2 = best_prev2_tags[i][t0][t1];
    t0 = t1;
    t1 = t2;
  }

  return best_scores[N-1][best_t.first][best_t.second];
}

////////////////////////////////////////////////////////////
// Top2

double Decoder::find_top2_diff(FeatureSet &features, Sequence &seq, bool use_averaged, int max_seg_len) {
  DoublePair scores = find_top2_scores(features, seq, use_averaged, max_seg_len);
  return scores.first - scores.second;
}

DoublePair Decoder::find_top2_scores(FeatureSet &features, Sequence &seq, bool use_averaged, int max_seg_len) {
  DoubleVec scores;
  Decoder decoder(features);

  if(features.type() == FTYPE_PT) {
    IntVecVec tagss;
    decoder.find_best_tags_beam_nbest(seq, tagss, scores, use_averaged, 2);
  }
  else if(features.type() == FTYPE_SEG) {
    SegVecVec segss;
    decoder.find_best_segs_dp_nbest(seq, segss, scores, use_averaged, max_seg_len, 2);
  }
  else
    assert(0);

  /*IntVec tags;
  double score = find_best_tags(seq, tags, use_averaged);

  track("find_top2_diff()", "", true);
  logs("DP seq " << score << ' ' << show_Sequence(D, seq, tags));
  logs("1st top seq " << scores[0] << ' ' << show_Sequence(D, seq, tagss[0]));
  logs("2nd top seq " << scores[1] << ' ' << show_Sequence(D, seq, tagss[1]));*/
  return DoublePair(scores[0], scores[1]);
}
