#include "basic/str.h"
#include "basic/hard-ofstream.h"
#include "basic/prob-utils.h"
#include "example-set.h"
#include "decoder.h"

extern bool print_wrong_details;
extern bool print_features_after_update;
extern bool print_all_errors;
extern bool eval_after_each_iter;
extern bool save_params;
extern int train_limit, test_limit;
extern int niters;
extern int max_seg_len;
extern string params_file, eval_file, stats_file;
extern bool use_bpm;

extern bool active_learning_sample_randomly;
extern bool active_learning_sample_margin;
extern bool active_learning_print_picked;
extern bool active_learning_norm_len;
extern double active_learning_incr;
extern double active_learning_ninit;

////////////////////////////////////////////////////////////

ExampleSet::ExampleSet(const string &tag_domain_str) : D(tag_domain_str), features(D) {
  D.boundary_token = Token::create_special_token(D, "$#$");
}

void ExampleSet::read(const char *train_file, const char *test_file, const char *test2_file, double train_frac) {
  read(train_file, train_sequences, train_frac);
  read(test_file, test_sequences, 1);
  if(test2_file && *test2_file)
    read(test2_file, test2_sequences, 1);
}

void ExampleSet::read(const char *file, SequenceVec &sequences, double keep_frac) {
  track("ExampleSet::read()", file, true);
  ifstream in(file);
  assert(in);
  read(in, sequences, keep_frac);
}

void ExampleSet::add_boundary(Sequence *seq) {
  foridx(i, boundary_size)
    seq->push_back(*D.boundary_token);
}

// Put a new sequence with a bounary marker in sequences and return it.
Sequence *ExampleSet::new_seq(SequenceVec &sequences) {
  Sequence *seq = &push_back(sequences);
  add_boundary(seq);
  return seq;
}

// Note: Pad the end of the sequence with the boundary marker.
Sequence *ExampleSet::flush_seq(Sequence *seq, SequenceVec &sequences) {
  if(seq_has_content(seq)) {
    add_boundary(seq);
    if(D.is_seg()) seq->set_segs(D);
    seq = new_seq(sequences);
  }
  return seq;
}

void ExampleSet::clear_seq(Sequence *seq) {
  seq->clear();
  add_boundary(seq);
}

// Uses global variables.
void ExampleSet::read(istream &in, SequenceVec &sequences, double keep_frac) {
  char buf[1024];

  int num_total_seqs = 0;
  Sequence *seq = new_seq(sequences);
  while(in.getline(buf, sizeof(buf))) {
    char *s = trim(buf);
    if(!*s) { // Just finished a sequence.
      if(seq_has_content(seq)) num_total_seqs++;

      if(rand_double() < keep_frac)
        seq = flush_seq(seq, sequences);
      else
        clear_seq(seq); // Throw the sequence away.
    }
    else if(!seq->push_back().parse(s, D)) // Add the token.
      seq->pop_back(); // Ignore invalid tokens.
  }
  if(seq_has_content(seq)) num_total_seqs++;
  flush_seq(seq, sequences);
  sequences.pop_back();

  logs("Read " << sequences.size() << '/' << num_total_seqs << " sequences.");
  logs(D.word_db.size() << " word types, " << D.pos_db.size() << " part-of-speech tags, "
    << D.chunk_db.size() << " chunk tags, " << D.ent_db.size() << " named-entity tags.");
}

int num_tokens(const SequenceVec &sequences, const BoolVec &use, const IntVec &perm) {
  int n = 0;
  foridx(i, len(sequences)) {
    if(len(use) > 0 && !use[i]) continue;
    const Sequence &seq = len(perm) > 0 ? sequences[perm[i]] : sequences[i];
    n += len(seq)-2*boundary_size;
  }
  return n;
}

// type is either segmentation or tagging.
// use_averaged: Use averaged weights in params.
// fix_weights: if there's an error, apply the perceptron update rule.
// use: only consider these examples (used in active learning).
Performance ExampleSet::process_sequences(FeatureType type, const char *name, SequenceVec &sequences,
    bool use_averaged, bool fix_weights, bool output_eval, const BoolVec &use, const IntVec &perm) {
  int eff_num_sequences = len(use) > 0 ? count_of(use, true) : len(sequences);
  track("process_sequences()", name << ": " << eff_num_sequences << '/' << len(sequences) << " sequences - " <<
        num_tokens(sequences, use, perm) << " tokens (" << feature_type_names[type] << ")", false);
  Performance perf(D);

  assert(len(use) == 0 || len(use) == len(sequences));
  assert(len(perm) == 0 || len(perm) == len(sequences));

  track_foridx(i, len(sequences), "Sequence", true) {
    if(len(use) > 0 && !use[i]) continue;
    Sequence &seq = len(perm) > 0 ? sequences[perm[i]] : sequences[i];

    // Decode.
    IntVec truth_tags, pred_tags;
    SegVec truth_segs, pred_segs;
    double truth_score = 0, pred_score = 0;
    bool correct = false;

    if(type == FTYPE_PT) {
      truth_tags = seq.extract_tags();
      truth_score = features.compute_score(seq, truth_tags, use_averaged);
      pred_score = Decoder::find_best_tags(features, seq, pred_tags, use_averaged);
      correct = (truth_tags == pred_tags);
      // If using floats, don't have enough precision for this to be equal.
      //assert_feq(pred_score, features.compute_score(seq, pred_tags, use_averaged));
    }
    if(type == FTYPE_SEG) {
      truth_segs = seq.extract_segs();
      truth_score = features.compute_score(seq, truth_segs, use_averaged);
      pred_score = Decoder::find_best_segs(features, seq, pred_segs, use_averaged, max_seg_len);
      correct = (truth_segs == pred_segs);
      //assert_feq(pred_score, features.compute_score(seq, pred_segs, use_averaged));
    }

    logs("truth score = " << truth_score);
    logs("pred score = " << pred_score);

    // Update stats.
    if(type == FTYPE_PT)  perf.update_stats(seq, truth_tags, pred_tags);
    if(type == FTYPE_SEG) perf.update_stats(seq, truth_segs, pred_segs);
    
    // Output the evaluation file.
    if(output_eval) {
      if(type == FTYPE_SEG) pred_tags = Sequence::segs_to_tags(D, pred_segs);
      forvec(j, const Token &, token, seq) {
        if(token.word == D.boundary_token->word) continue;
        eval_out << D.word_db[token.word] << ' ' << D.tag_db[token.tag] << ' ' << D.tag_db[pred_tags[j]] << endl;
      }
      eval_out << endl;
    }

    // Print error.
    if(!correct) {
      if(type == FTYPE_PT)  logs("Wrong: " << show_Sequence(D, seq, pred_tags));
      if(type == FTYPE_SEG) logs("Wrong: " << show_Sequence(D, seq, Sequence::segs_to_tags(D, pred_segs)));

      if(print_wrong_details) {
        if(type == FTYPE_PT)  features.print_tag_diff(seq, truth_tags, pred_tags, use_averaged);
        if(type == FTYPE_SEG) features.print_seg_diff(seq, truth_segs, pred_segs, use_averaged);
      }
      perf.errors.push_back(i);
    }

    // Update weights.
    if(fix_weights) {
      if(type == FTYPE_PT)  features.update_weights(seq, pred_tags);
      if(type == FTYPE_SEG) features.update_weights(seq, pred_segs);

      if(!correct) {
        if(print_features_after_update) {
          logs(perf);
          perf.print_confusion();
        }
      }
    }

    logs(prog_status);
  }
  log_info.hard_flush();

  return perf;
}

struct IterationPerformance {
  IterationPerformance(int iter, const Domain &D) : iter(iter), training(D), validation(D), test(D), test2(D) { }

  int iter;
  Performance training; // While training
  Performance validation; // Validation set (using averaged weights)
  Performance test; // Test set (using averaged weights)
  Performance test2; // Test2 set (using averaged weights)
};

// Pick the n examples that are most uncertain, i.e., the ones that have the smallest
// u-margin, but take the number of tokens in the sequence into consideration.
int ExampleSet::pick_examples(BoolVec &in_set, int n, bool pick_randomly) {
  int N = len(train_sequences);
  n <?= N - count_of(in_set, true);

  track("pick_examples()", "picking " << n, false);

  if(pick_randomly) {
    foridx(_, n) {
      int i;
      do { i = rand() % N; } while(in_set[i]);
      in_set[i] = true;
    }
  }
  else {
    // Compute u-margins of all training examples not in the set.
    DoubleIntVec umargins;
    BigStatFig fig;
    track_block("Measuring u-margins", "", false) {
      track_forvec(i, Sequence &, seq, train_sequences, "Measuring u-margin of sequence", true) {
        if(!in_set[i]) {
          // Find out more about these u-margin measures.
#if 0
          DoubleVec scores;
          Decoder decoder(features);

          IntVec tags;
          if(features.type() == FTYPE_PT) {
            IntVecVec tagss;
            decoder.find_best_tags_beam_nbest(seq, tagss, scores, true, 2);
            tags = tagss[0];
          }
          else if(features.type() == FTYPE_SEG) {
            SegVecVec segss;
            decoder.find_best_segs_dp_nbest(seq, segss, scores, true, max_seg_len, 2);
            tags = Sequence::segs_to_tags(D, segss[0]);
          }
          double accuracy = 0;
          foridx(j, len(tags))
            if(tags[j] == seq[j].tag) accuracy++;
          accuracy /= len(tags);

          double umargin = scores[0] - scores[1];
          cerr << n << ' ' << scores[0] << ' ' << scores[1] << ' ' << umargin << ' ' << len(seq)-2*boundary_size << ' ' << accuracy << endl;
#else
          double umargin = Decoder::find_top2_diff(features, seq, true, max_seg_len);
#endif

          if(active_learning_norm_len) umargin *= len(seq);
          umargins.push_back(DoubleInt(umargin, i));
          fig.add(umargin);
          logs("U-margin = " << umargin << ", length = " << len(seq) << ", seq " << i);
        }
      }
    }

    logss("All u-margins: " << fig);

    fig.clear();
    assert(n <= len(umargins));

    if(active_learning_sample_margin) {
      track_block("Sample n based on u-margin", n, true) {
        DoubleVec probs(len(umargins));

        foridx(i, len(umargins)) {
          double m = umargins[i].first >? 1; // If margin screwed up... (MAKE SURE 1 is a small number for a margin)
          probs[i] = 1 / m;
        }
        norm_distrib(probs);

        foridx(q, n) {
          // Pick an example that's not in the set already
          int i;
          do {
            i = rand_index(probs);
            if(!in_set[umargins[i].second]) break;
            probs[i] = 0; // Got one already picked, so kill it's probability so it doesn't happen again.
            norm_distrib(probs);
          } while(1);

          fig.add(umargins[i].first);
          in_set[umargins[i].second] = true;

          if(active_learning_print_picked) {
            const Sequence &seq = train_sequences[umargins[i].second];
            logs("U-margin = " << umargins[i].first << ", sentence " << umargins[i].second <<
                 ' ' << show_Sequence(D, seq, seq.extract_tags()));
          }
        }
      }
      logss("Sampled u-margins: " << fig);
    }
    else {
      // Pick the top n.
      track_block("Pick top n", n, true) {
        fig.clear();
        assert(n <= len(umargins));

        sort(umargins.begin(), umargins.end(), less<DoubleInt>());
        foridx(i, n) {
          fig.add(umargins[i].first);
          in_set[umargins[i].second] = true;

          if(active_learning_print_picked) {
            const Sequence &seq = train_sequences[umargins[i].second];
            logs("U-margin = " << umargins[i].first << ", sentence " << umargins[i].second <<
                 ' ' << show_Sequence(D, seq, seq.extract_tags()));
          }
        }
      }
      logss("Top u-margins: " << fig);
    }
  }

  return n;
}

int ExampleSet::num_words(const SequenceVec &sequences, const BoolVec &use) {
  int n = 0;
  forvec(i, const Sequence &, seq, sequences)
    if(use[i])
      n += len(seq) - boundary_size*2;
  return n;
}

void ExampleSet::train_perceptron_with_active_learning() {
  track("train_perceptron_with_active_learning()", "", false);

  int ntotal_examples = len(train_sequences);
  BoolVec in_train_set(ntotal_examples); // Examples in our current training set

  hard_ofstream stats_out(stats_file.c_str());

  FeatureType type = features.type();
  int nexamples = 0;
  while(nexamples < ntotal_examples) {
    track("", "Currently using " << nexamples << '/' << ntotal_examples << " training examples", false);

    // Pick some examples.
#define getn(x, y) int((x) < 1 ? ((x)*(y)+0.5) : (x))
    int nnew;
    if(nexamples == 0) nnew = getn(active_learning_ninit, len(train_sequences));
    else               nnew = getn(active_learning_incr, nexamples);

    nexamples += pick_examples(in_train_set, nnew,
        active_learning_sample_randomly || nexamples == 0);

#if 0
    if(nexamples == 0) {
      ifstream in(active_learning_seed_file.c_str());
      assert2(in, active_learning_seed_file);

      int i;
      while(in >> i) {
        assert(i >= 0 && i <= ntotal_examples && !in_train_set[i]);
        in_train_set[i] = true;
        nexamples++;
      }

      // Initial seed
      int n = active_learning_incr <? ntotal_examples;
      foridx(i, n)
        in_train_set[i] = true;
      nexamples += n;
    }
    else {
      nexamples += pick_examples(in_train_set, 100, active_learning_sample_randomly);
    }
#endif

    // Train
    track_block("Train", "", false) {
      features.clear_weights();
      for(int iter = 1; iter <= niters; iter++) {
        track("Iteration", iter << '/' << niters, true);

        Performance training_perf = process_sequences(type, "Training", train_sequences, false, true, false, in_train_set, IntVec());
        logs("While training: " << training_perf);
      }
    }

    // Test
    Performance test_perf = process_sequences(type, "Test (averaged)", test_sequences, true, false, false, BoolVec(), IntVec());
    int nwords = num_words(train_sequences, in_train_set);
    stats_out << nexamples << ' ' << nwords << ' ' << test_perf.word_error.val() << ' ' << test_perf.F1() << endl;
    stats_out.hard_flush();
    logss("Test (trained with " << nexamples << " sequences): " << test_perf); test_perf.print_confusion();
  }
}

int ExampleSet::compute_max_seg_len(const SequenceVec &sequences) {
  int max_n = 0;
  forvec(_, const Sequence &, seq, sequences) {
    SegVec segs = seq.extract_segs();
    forvec(_, Seg, seg, segs) {
      if(seg.a == -1) continue;
      max_n >?= seg.n;
    }
  }
  return max_n;
}

void ExampleSet::train_perceptron() {
  track("train_perceptron()", "", false);

  FeatureType type = features.type();

  // Stats file
  hard_ofstream stats_out;
  if(!stats_file.empty()) {
    stats_out.open(stats_file.c_str());
    assert2(stats_out, stats_file);
  }

  logs(len(train_sequences) << " sequences, " << niters << " iterations.");

  vector<IterationPerformance> iperfs; // iteration-1 -> performance on that iteration

  for(int iter = 1; iter <= niters; iter++) {
    track("Iteration", iter << '/' << niters, true);

    IterationPerformance &iperf = push_back(iperfs, IterationPerformance(iter, D));

    if(use_bpm) {
      features.clear_plain_weights();
      features.begin_update();
    }

    iperf.training = process_sequences(type, "Training", train_sequences, false, true, false, BoolVec(),
        use_bpm ? rand_permutation(len(train_sequences)) : IntVec());
    logs("While training: " << iperf.training);

    if(use_bpm) features.end_update();

    // Evaluate performance.
    if(eval_after_each_iter || iter == niters) {
      // On validation.
      iperf.validation = process_sequences(type, "Validation (averaged)", validation_sequences, true, false, false, BoolVec(), IntVec());
      logss("Validation (averaged): " << iperf.validation); iperf.validation.print_confusion();

      // On test.
      eval_out.open(eval_file.c_str());
      iperf.test = process_sequences(type, "Test (averaged)", test_sequences, true, false, true, BoolVec(), IntVec());
      eval_out.close();
      logss("Test (averaged): " << iperf.test); iperf.test.print_confusion();

      // On test2.
      iperf.test2 = process_sequences(type, "Test2 (averaged)", test2_sequences, true, false, false, BoolVec(), IntVec());
      logss("Test2 (averaged): " << iperf.test2); iperf.test2.print_confusion();
    }

    // Print to stats file.
    if(stats_out) {
      stats_out << iter << ' '
                << iperf.training.word_error.val() << ' '
                << iperf.validation.word_error.val() << ' '
                << iperf.test.word_error.val();
      if(D.is_seg()) {
        stats_out << ' '
                  << iperf.training.F1() << ' '
                  << iperf.validation.F1() << ' '
                  << iperf.test.F1() << ' '
                  << iperf.test.precision.val() << ' '
                  << iperf.test.recall.val() << ' '
                  << iperf.test2.F1() << ' '
                  << iperf.test2.precision.val() << ' '
                  << iperf.test2.recall.val();
      }
      stats_out << endl;
      stats_out.hard_flush();
    }
  }


  // Print all errors.
  if(print_all_errors) {
    const IterationPerformance &iperf = iperfs[len(iperfs)-1];
    track("Test errors", len(iperf.test.errors) << " errors", true);
    forvec(_, int, i, iperf.test.errors) {
      Sequence &seq = test_sequences[i];

      // Decode.
      IntVec truth_tags, pred_tags;
      SegVec truth_segs, pred_segs;
      double truth_score = 0, pred_score = 0;

      if(type == FTYPE_PT) {
        truth_tags = seq.extract_tags();
        truth_score = features.compute_score(seq, truth_tags, true);
        pred_score = Decoder::find_best_tags(features, seq, pred_tags, true);
      }
      if(type == FTYPE_SEG) {
        truth_segs = seq.extract_segs();
        truth_score = features.compute_score(seq, truth_segs, true);
        pred_score = Decoder::find_best_segs(features, seq, pred_segs, true, max_seg_len);
      }

      logs("pred score = " << pred_score);
      logs("truth score = " << truth_score);

      // Print error.
      if(type == FTYPE_PT)  logs(show_Sequence(D, seq, pred_tags));
      if(type == FTYPE_SEG) logs(show_Sequence(D, seq, Sequence::segs_to_tags(D, pred_segs)));
      if(print_wrong_details) {
        if(type == FTYPE_PT)  features.print_tag_diff(seq, truth_tags, pred_tags, true);
        if(type == FTYPE_SEG) features.print_seg_diff(seq, truth_segs, pred_segs, true);
      }
    }
  }

  // Process stats file.
  if(stats_out) {
    stats_out.close();

    string dir = get_dir(stats_file);
    string gnuplot_file = stats_file+".gnuplot";

    ofstream out(gnuplot_file.c_str());
    out << "set term postscript eps color\n";

    out << "set output '" << dir+"/stats-error.eps" << "'\n";
    out << "plot \\\n";
    out << "  '" << stats_file << "' using 1:2 with lines title 'Error while training', \\\n";
    out << "  '" << stats_file << "' using 1:3 with lines title 'Validation (averaged) error', \\\n";
    out << "  '" << stats_file << "' using 1:4 with lines title 'Test (averaged) error'\n";

    if(D.is_seg()) {
      out << "set output '" << dir+"/stats-f1.eps" << "'\n";
      out << "plot \\\n";
      out << "  '" << stats_file << "' using 1:5 with lines title 'F1 while training', \\\n";
      out << "  '" << stats_file << "' using 1:6 with lines title 'Validation (averaged) F1',\\\n";
      out << "  '" << stats_file << "' using 1:7 with lines title 'Test (averaged) F1'\n";
      out.close();
    }

    if(len(iperfs) > 1)
      system(("gnuplot " + gnuplot_file).c_str());
  }

  // Find the iteration that gave the best performance on the validation set.
  DoubleInt best_valid_error = DoubleInt(DBL_MAX, -1), best_valid_F1 = DoubleInt(DBL_MIN, -1);
  forvec(i, const IterationPerformance &, iperf, iperfs) {
    improve(best_valid_error, DoubleInt(iperf.validation.word_error.val(), i), less<double>());
    improve(best_valid_F1, DoubleInt(iperf.validation.F1(), i), greater<double>());
  }

  if(best_valid_error.second != -1) {
    const IterationPerformance &error_iperf = iperfs[best_valid_error.second];
    logs("Best validation error is at iteration " << error_iperf.iter << '/' << len(iperfs)
                                               << ": " << error_iperf.validation.word_error
                                               << ", test error = " << error_iperf.test.word_error);
  }
  if(best_valid_F1.second != -1) {
    const IterationPerformance &F1_iperf = iperfs[best_valid_F1.second];
    logs("Best validation F1 is at iteration " << F1_iperf.iter << '/' << len(iperfs)
                                               << ": " << F1_iperf.validation.F1()
                                               << ", test F1 = " << F1_iperf.test.F1());
  }
}

// Just take the last few training examples for validation.
void ExampleSet::steal_train_data_for_cross_validation(double ratio) {
  int N = len(train_sequences);
  int n = int(N * ratio);
  track("steal_train_data_for_cross_validation", ratio << " = " << n << '/' << N, false);

  assert(len(validation_sequences) == 0);

  for(int i = N-n; i < N; i++)
    validation_sequences.push_back(train_sequences[i]);
  train_sequences.resize(N-n);
}

void ExampleSet::precompute_features() {
  precompute_features(train_sequences);
  precompute_features(validation_sequences);
  precompute_features(test_sequences);
  precompute_features(test2_sequences);
}

void ExampleSet::precompute_features(SequenceVec &sequences) {
  track("ExampleSet::precompute_features()", len(sequences), false);
  track_forvec(_, Sequence &, seq, sequences, "Sequence", true)
    features.precompute_features(seq);
}

void ExampleSet::finish_initialization() {
  // Take only first {train_test}_limit sequences.
  if(train_limit < len(train_sequences)) train_sequences.resize(train_limit);
  if(test_limit < len(test_sequences)) test_sequences.resize(test_limit);

  if(max_seg_len == -1) max_seg_len = compute_max_seg_len(train_sequences)+1;
}
