#include "basic/stl-utils.h"
#include "basic/stats.h"
#include "basic/str.h"
#include "basic/mem-tracker.h"
#include "feature-set.h"
#include "token.h"

extern string clusters_file;
extern string strength_file;
extern string map_file;
extern bool precompute_features;
extern bool use_bpm;
extern bool use_cache;
extern int max_seg_len;
extern int nprefix_incr;

////////////////////////////////////////////////////////////

TransformMapSet::TransformMapSet() {
  word_form_map = brief_word_form_map = word_lowercase_map = NULL;
}

FeatureSet::FeatureSet(Domain &D) : D(D) {
  clusters_loaded = false;
  strength_loaded = false;
  map_loaded = false;
}

FeatureSet::~FeatureSet() {
  forvec(_, FeatureTemplate *, ft, all_templates) delete ft;
  forvec(_, TransformMap *, map, all_maps) delete map;
  forvec(_, StringTransformer *, transformer, all_transformers) delete transformer;
}

FeatureType FeatureSet::type() const {
  assert(len(templates) > 0);
  FeatureType t = FTYPE_PT;
  forvec(i, FeatureTemplate *, ft, templates) {
    if(i == 0) t = ft->type();
    else assert(t == ft->type());
  }
  return t;
}

////////////////////////////////////////////////////////////

// The children are ints.
IntVec get_ints(const LispNode *node) {
  IntVec ints;
  forvec(_, const LispNode *, subnode, node->children) {
    assert(len(subnode->children) == 0);
    ints.push_back(atoi(subnode->value.c_str()));
  }
  return ints;
}

// The children are doubles.
DoubleVec get_doubles(const LispNode *node) {
  DoubleVec doubles;
  forvec(_, const LispNode *, subnode, node->children) {
    assert(len(subnode->children) == 0);
    doubles.push_back(atof(subnode->value.c_str()));
  }
  return doubles;
}

// The children are int vectors.
IntVecVec get_int_vecs(const LispNode *node) {
  IntVecVec int_vecs;
  forvec(_, const LispNode *, subnode, node->children) {
    assert(len(subnode->children) == 0);
    StringVec strings = split(subnode->value.c_str(), ",", true);
    IntVec ints(len(strings));
    foridx(i, len(strings)) ints[i] = atoi(strings[i].c_str());
    int_vecs.push_back(ints);
  }
  return int_vecs;
}

void FeatureSet::parse_map(const LispNode *node, vector<TransformMap *> &maps, StrDB &db, TransformMapSet &set) {
  if(node->value == "identity")
    maps.push_back(NULL);
  else if(node->value == "suffix") {
    IntVec ints = get_ints(node);
    forvec(_, int, k, ints) {
      TransformMap *map = set.word_suffix_map[k];
      if(!map) map = set.word_suffix_map[k] = mem(new TransformMap(mem(new T_Suffix(k)), db));
      maps.push_back(map);
    }
  }
  else if(node->value == "prefix") {
    IntVec ints = get_ints(node);
    forvec(_, int, k, ints) {
      TransformMap *map = set.word_prefix_map[k];
      if(!map) map = set.word_prefix_map[k] = mem(new TransformMap(mem(new T_Prefix(k)), db));
      maps.push_back(map);
    }
  }
  else if(node->value == "lowercase") {
    if(!set.word_lowercase_map) set.word_lowercase_map = mem(new TransformMap(mem(new T_Lowercase()), db));
    maps.push_back(set.word_lowercase_map);
  }
  else if(node->value == "crypt") {
    IntVec ints = get_ints(node);
    assert(len(ints) == 2);
    int nkeys = ints[0], range = ints[1];
    foridx(key, nkeys) {
      TransformMap *map = set.crypt_maps[key];
      if(!map)
        map = set.crypt_maps[key] = mem(new TransformMap(mem(new T_Crypt(key, range)), db));
      maps.push_back(map);
    }
  }
  else if(node->value == "form") {
    if(!set.word_form_map) set.word_form_map = mem(new TransformMap(mem(new T_Form(false)), db));
    if(!set.brief_word_form_map) set.brief_word_form_map = mem(new TransformMap(mem(new T_Form(true)), db));
    maps.push_back(set.word_form_map);
    maps.push_back(set.brief_word_form_map);
  }
  else if(node->value == "contains") {
    forvec(_, const LispNode *, subnode, node->children) {
      const string &descrip = subnode->value;
      TransformMap *map = set.word_contains_map[descrip];
      if(!map) map = set.word_contains_map[descrip] = mem(new TransformMap(mem(new T_Contains(descrip)), db));
      maps.push_back(map);
    }
  }
  else if(node->value == "cluster_path") {
    if(!clusters_loaded) {
      track("Loading clusters", clusters_file, false);
      // File could be directory as well, in which case, we load all the clusters
      vector<string> files;
      if(!get_files_in_dir(clusters_file, true, files))
        files.push_back(clusters_file);

      logs(len(files) << " files");

      clusters_dbs.resize(len(files));
      clusters_names.resize(len(files));
      forvec(i, const string &, file, files) {
        clusters_dbs[i].read(file.c_str());
        clusters_names[i] = strip_dir(file_base(file));
      }
      clusters_loaded = true;
    }

    int nc = len(clusters_dbs);
    forvec(c, StrStrDB &, clusters_db, clusters_dbs) {
      for(int k = 0; k < clusters_db.max_t_len; k += nprefix_incr) {
        string name = clusters_names[c]+string(clusters_names[c].empty()?"":"-")+"cluster_path";
        TransformMap *map = set.word_cluster_path_map[k*nc+c];
        if(!map) {
          StringTransformer *transformer = mem(new T_TableLookup(name.c_str(), clusters_db));
          if(k > 0) // Prefix
            transformer = mem(new ComposeStringTransformer(transformer, mem(new T_Prefix(k))));
          map = set.word_cluster_path_map[k*nc+c] = mem(new TransformMap(transformer, db));
        }
        maps.push_back(map);
      }
    }
  }
}

vector<TransformMap *> FeatureSet::get_maps(const LispNode *node, StrDB &db, TransformMapSet &set) {
  vector<TransformMap *> maps;
  forvec(_, const LispNode *, subnode, node->children)
    parse_map(subnode, maps, db, set);
  return maps;
}

void FeatureSet::parse_feature_templates(const LispNode *node, vector<FeatureTemplate *> &templates) {
  int n = len(node->children);

  // Tag features
  if(node->value == "CurrTag") {
    assert2(n == 0, node->value);
    templates.push_back(mem(new F_CurrTag(D)));
    dp_order >?= 1;
  }
  else if(node->value == "PrevTag_CurrTag") {
    assert2(n == 0, node->value);
    templates.push_back(mem(new F_PrevTag_CurrTag(D)));
    dp_order >?= 2;
  }
  else if(node->value == "Prev2Tag_PrevTag_CurrTag") {
    assert2(n == 0, node->value);
    templates.push_back(mem(new F_Prev2Tag_PrevTag_CurrTag(D)));
    dp_order >?= 3;
  }
  else if(node->value == "Word") {
    assert2(n == 2, node->value << " takes 2 arguments; got " << n);
    IntVecVec offsets = get_int_vecs(node->children[0]);
    vector<TransformMap *> maps = get_maps(node->children[1], D.word_db, word_maps);

    forvec(_, TransformMap *, map, maps) {
      forvec(_, const IntVec &, offset, offsets) {
        FeatureTemplate *ft = NULL;
        forvec(i, int, o, offset) {
          FeatureTemplate *sub_ft = mem(new F_Word(D, o, map));
          if(ft) ft = mem(product_ft(ft, sub_ft, i == 1 ? &bigram_db : NULL)); // Use the bigram (word, word -> bigram) db if possible
          else   ft = sub_ft;
        }
        assert(ft);
        if(use_cache) ft = cache(ft);
        templates.push_back(ft);
      }
    }
  }
  else if(node->value == "Words") {
    assert2(n == 2, node->value << " takes 2 arguments; got " << n);
    IntVecVec offsets = get_int_vecs(node->children[0]);
    vector<TransformMap *> maps = get_maps(node->children[1], D.phrase_db, phrase_maps);

    forvec(_, TransformMap *, map, maps) {
      forvec(_, const IntVec &, offset, offsets) {
        assert(len(offset) == 2);
        int o = offset[0], n = offset[1]-offset[0]+1;
        FeatureTemplate *ft = mem(new F_Words(D, o, n, map));
        if(use_cache) ft = cache(ft);
        templates.push_back(ft);
      }
    }
  }
  else if(node->value == "Map") {
    assert2(n == 1, node->value << " takes 1 argument; got " << n);
    IntVecVec offsets = get_int_vecs(node->children[0]);

    if(!map_loaded) {
      map_db.read(map_file.c_str(), &D.phrase_db);
      //map_db.normalize();
      map_loaded = true;
    }

    forvec(_, const IntVec &, offset, offsets) {
      assert(len(offset) == 2);
      int o = offset[0], n = offset[1]-offset[0]+1;
      FeatureTemplate *ft = mem(new F_Map(D, o, n, map_db));
      if(use_cache) ft = cache(ft);
      templates.push_back(ft);
    }
  }
  else if(node->value == "POS") {
    assert2(n == 1, node->value << " takes 1 argument; got " << n);
    IntVecVec offsets = get_int_vecs(node->children[0]);

    forvec(_, const IntVec &, offset, offsets) {
      FeatureTemplate *ft = NULL;
      forvec(_, int, o, offset) {
        FeatureTemplate *sub_ft = mem(new F_POS(D, o));
        if(ft) ft = mem(product_ft(ft, sub_ft, NULL));
        else   ft = sub_ft;
      }
      assert(ft);
      if(use_cache) ft = cache(ft);
      templates.push_back(ft);
    }
  }
  else if(node->value == "Strength") {
    assert2(n <= 2, node->value << " takes 0-2 arguments; got " << n);
    double scale = 1, power = 1;
    if(n >= 1) scale = atof(node->children[0]->value.c_str());
    if(n >= 2) power = atof(node->children[1]->value.c_str());

    if(!strength_loaded) {
      strength_db.read(strength_file.c_str(), D.word_db);
      //strength_db.normalize();
      strength_loaded = true;
    }

    foridx(i, strength_db.nstrengths) {
      FeatureTemplate *ft = mem(new F_Strength(D, strength_db, scale, power, i));
      if(use_cache) ft = cache(ft);
      templates.push_back(ft);
    }
  }
  // Threshold
  else if(node->value == "Threshold") {
    assert2(n == 2, node->value << " takes 2 arguments (threshold(start,skip,stop), feature templates); got " << n);

    DoubleVec doubles = get_doubles(node->children[0]);
    assert2(len(doubles) == 3, "start, skip, stop (inclusive)");

    double start = doubles[0], skip = doubles[1], stop = doubles[2];

    vector<FeatureTemplate *> sub_templates;
    parse_feature_templates(node->children[1], sub_templates);

    forvec(_, FeatureTemplate *, sub_ft, sub_templates) {
      assert(sub_ft->is_single() && !sub_ft->is_bool());

      for(double threshold = start; threshold <= stop; threshold += skip) {
        FeatureTemplate *ft = mem(new ThresholdSingleBoolFeatureTemplate(*(SingleFeatureTemplate *)sub_ft, threshold));
        if(use_cache) sub_ft = cache(sub_ft);
        templates.push_back(ft);
      }
    }
  }
  // Segment features
  else if(node->value == "CurrSeg") {
    assert2(n == 0, node->value);
    templates.push_back(mem(new F_CurrSeg(D)));
  }
  else if(node->value == "PrevSeg_CurrSeg") {
    assert2(n == 0, node->value);
    templates.push_back(mem(new F_PrevSeg_CurrSeg(D)));
  }
  else if(node->value == "Phrase") {
    assert2(n == 1, node->value << " takes 1 argument; got " << n);
    vector<TransformMap *> maps = get_maps(node->children[0], D.phrase_db, phrase_maps);

    forvec(_, TransformMap *, map, maps) {
      FeatureTemplate *ft = mem(new F_Phrase(D, map));
      if(use_cache) ft = cache(ft);
      templates.push_back(ft);
    }
  }
  else if(node->value == "POSPhrase") {
    assert2(n == 0, node->value << " takes 0 argument; got " << n);
    FeatureTemplate *ft = mem(new F_POSPhrase(D));
    if(use_cache) ft = cache(ft);
    templates.push_back(ft);
  }
  else if(node->value == "FocusPhrase") {
    assert2(n == 2, node->value << " takes 2 arguments; got " << n);
    Focus focus = parse_focus(node->children[0]->value);

    vector<FeatureTemplate *> sub_templates;
    parse_feature_templates(node->children[1], sub_templates);

    forvec(_, FeatureTemplate *, sub_ft, sub_templates) {
      //if(use_cache) sub_ft = cache(sub_ft); NO NO, DUPLICATE!
      FeatureTemplate *ft =
        sub_ft->is_single() && sub_ft->is_bool() ?
          mem(new F_SingleBoolFocusPhrase(*(SingleBoolFeatureTemplate *)sub_ft, focus)) :
          mem(new F_FocusPhrase(*(MultiFeatureTemplate *)sub_ft, focus));
      templates.push_back(ft);
    }
  }
  else if(node->value == "AggregatePhrase") {
    assert2(n == 2, node->value << " takes 2 arguments; got " << n);

    int start = atoi(node->children[0]->value.c_str());

    vector<FeatureTemplate *> sub_templates;
    parse_feature_templates(node->children[1], sub_templates);

    forvec(_, FeatureTemplate *, sub_ft, sub_templates) {
      //if(use_cache) sub_ft = cache(sub_ft); NO NO, DUPLICATE!
      FeatureTemplate *ft = mem(new F_AggregatePhrase(*(MultiFeatureTemplate *)sub_ft, start));
      templates.push_back(ft);
    }
  }
  else if(node->value == "PhraseLength") {
    assert2(n == 0, node->value << " takes 0 arguments; got " << n);
    FeatureTemplate *ft = mem(new F_PhraseLength(D));
    if(use_cache) ft = cache(ft);
    templates.push_back(ft);
  }
  else if(node->value == "PhraseStrength") {
    assert2(n <= 2, node->value << " takes 0-2 arguments; got " << n);
    double scale = 1, power = 1;
    if(n >= 1) scale = atof(node->children[0]->value.c_str());
    if(n >= 2) power = atof(node->children[1]->value.c_str());

    if(!strength_loaded) {
      strength_db.read(strength_file.c_str(), D.word_db);
      //strength_db.normalize();
      strength_loaded = true;
    }

    foridx(i, strength_db.nstrengths) {
      FeatureTemplate *ft = mem(new F_PhraseStrength(D, strength_db, scale, power, i));
      if(use_cache) ft = cache(ft);
      templates.push_back(ft);
    }
  }
  else
    assert2(false, "Unknown feature template: " << node->value);
}

void FeatureSet::read_templates(const char *file) {
  track("FeatureSet::read_templates()", file, true);

  LispTree tree;
  tree.read(file);

  dp_order = 0;
  forvec(_, const LispNode *, node, tree.root->children) {
    // For each set of feature templates...
    const vector<LispNode *> &children = node->children;
    int n = len(children);
    assert(n >= 1);

    // Each set of feature templates is a product between the feature templates specified by the children.
    vector<FeatureTemplate *> curr_templates;
    parse_feature_templates(node->children[n-1], curr_templates);
    for(int i = n-2; i >= 0; i--) {
      vector<FeatureTemplate *> sub_templates;
      parse_feature_templates(node->children[i], sub_templates);

      IntPairIntDB *f_db = NULL;
      vector<FeatureTemplate *> new_templates;
      forvec(_, FeatureTemplate *, ft1, sub_templates) {
        forvec(_, FeatureTemplate *, ft2, curr_templates) {
          FeatureTemplate *ft;
          ft = product_ft(ft1, ft2, f_db);
          f_db = ft->f_db; // Careful!  Sharing f_db's across these feature templates
          new_templates.push_back(mem(ft));
        }
      }
      curr_templates = new_templates;
    }

    forvec(_, FeatureTemplate *, ft, curr_templates)
      templates.push_back(ft);
  }

  track_block("", len(templates) << " feature templates", true) {
    forvec(_, FeatureTemplate *, ft, templates) {
      logs(ft->description());
    }
  }
  logs(len(cached_templates) << " cached feature templates");
}

bool same_feature_values(const IntDoubleVec &fvs1, const IntDoubleVec &fvs2) {
  if(len(fvs1) != len(fvs2)) return false;
  foridx(i, len(fvs1)) {
    const IntDouble &fv1 = fvs1[i], &fv2 = fvs2[i];
    if(fv1.first != fv2.first || !feq(fv1.second, fv2.second)) return false;
  }
  return true;
}

// Print the difference between feature values between truth_tags and pred_tags.
void FeatureSet::print_tag_diff(Sequence &seq, const IntVec &truth_tags, const IntVec &pred_tags, bool use_averaged) {
  track("print_tag_diff()", "", true);
  PtContext truth_context(D, seq, use_averaged);
  PtContext pred_context(D, seq, use_averaged);

  double score_diff = 0;
  foridx(i, len(seq)) {
    truth_context.step(truth_tags[i]);
    pred_context.step(pred_tags[i]);

    if(seq[i].word == D.boundary_token->word) continue;

    forvec(_, FeatureTemplate *, ft, templates) {
      IntDoubleVec truth_feature_values, pred_feature_values;
      ft->get_feature_values(truth_context, truth_feature_values);
      ft->get_feature_values(pred_context, pred_feature_values);

      if(same_feature_values(truth_feature_values, pred_feature_values)) continue;

      double truth_score = ft->score(truth_context);
      double pred_score = ft->score(pred_context);
      score_diff += pred_score - truth_score;

      if(len(truth_feature_values) != 1 || len(pred_feature_values) != 1) {
        logs("Truth: " << "i=" << i << ", score = " << ft->score(truth_context));
        forvec(_, const IntDouble &, fv, truth_feature_values)
          logs("  " << ft->description(fv.first) << " = " << fv.second);
        logs("Pred: " << "i=" << i << ", score = " << ft->score(pred_context));
        forvec(_, const IntDouble &, fv, pred_feature_values)
          logs("  " << ft->description(fv.first) << " = " << fv.second);
      }
      else {
        IntDouble tfv = truth_feature_values[0];
        IntDouble pfv = pred_feature_values[0];
        logs("Truth: " << "i=" << i << ", " << ft->description(tfv.first) << ", score = " << ft->score(truth_context));
        logs("Pred:  " << "i=" << i << ", " << ft->description(pfv.first) << ", score = " << ft->score(pred_context));
      }
    }
  }

  logs("Pred score - truth score = " << score_diff);
}

// Print the difference between feature values between truth_segs and pred_segs.
void FeatureSet::print_seg_diff(Sequence &seq, const SegVec &truth_segs, const SegVec &pred_segs, bool use_averaged) {
  track("print_seq_diff()", "", true);
  SegContext truth_context(D, seq, use_averaged);
  SegContext pred_context(D, seq, use_averaged);

  double score_diff = 0;
  foridx(i, len(seq)) {
    bool truth_exists = truth_segs[i].a != -1;
    bool pred_exists = pred_segs[i].a != -1;

    if(seq[i].word == D.boundary_token->word) continue;

    if(truth_exists) { truth_context.step(truth_segs[i]); assert(truth_context.i == i); }
    if(pred_exists) { pred_context.step(pred_segs[i]); assert(pred_context.i == i); }

    forvec(_, FeatureTemplate *, ft, templates) {
      IntDoubleVec truth_feature_values, pred_feature_values;
      ft->get_feature_values(truth_context, truth_feature_values);
      ft->get_feature_values(pred_context, pred_feature_values);

      if(truth_feature_values == pred_feature_values) continue;

      forvec(_, const IntDouble &, fv, truth_feature_values) {
        double truth_score = ft->score(truth_context);
        logs("Truth: " << "i=" << i << ", " << ft->description(fv.first) << ", score = " << ft->score(truth_context));
        score_diff -= truth_score;
      }

      forvec(_, const IntDouble &, fv, pred_feature_values) {
        double pred_score = ft->score(pred_context);
        logs("Pred:  " << "i=" << i << ", " << ft->description(fv.first) << ", score = " << ft->score(pred_context));
        score_diff += pred_score;
      }
    }
  }

  logs("Pred score - truth score = " << score_diff);
}

void FeatureSet::clear_weights() {
  forvec(_, FeatureTemplate *, ft, templates) ft->params.clear_weights();
}

// Update the weight parameters by incr given tag sequence tags.
void FeatureSet::update_weights_using_tags(Sequence &seq, const IntVec &tags, double incr) {
  PtContext context(D, seq, false);

  foridx(i, len(seq)) {
    context.step(tags[i]);
    if(seq[i].word == D.boundary_token->word) continue; // Skip boundary.

    forvec(_, FeatureTemplate *, ft, templates)
      ft->update_weights(context, incr);
  }
}

// Update the weight parameters by incr given a segmentation.
void FeatureSet::update_weights_using_segs(Sequence &seq, const SegVec &segs, double incr) {
  SegContext context(D, seq, false);

  foridx(i, len(seq)) {
    if(segs[i].a == -1) continue;
    context.step(segs[i]);
    if(seq[i].word == D.boundary_token->word) continue; // Skip boundary.

    forvec(_, FeatureTemplate *, ft, templates)
      ft->update_weights(context, incr);
  }
}

void FeatureSet::update_weights(Sequence &seq, const IntVec &pred_tags) {
  track("update_weights()", "", false);
  if(!use_bpm) begin_update();

  update_weights_using_tags(seq, seq.extract_tags(), 1); // True
  update_weights_using_tags(seq, pred_tags,         -1); // Pred

  if(!use_bpm) end_update();
}

void FeatureSet::update_weights(Sequence &seq, const SegVec &pred_segs) {
  track("update_weights()", "", false);
  if(!use_bpm) begin_update();

  update_weights_using_segs(seq, seq.extract_segs(), 1); // True
  update_weights_using_segs(seq, pred_segs,         -1); // Pred

  if(!use_bpm) end_update();
}

// Compute the score (sum of feature values * weights for all features based on context).
double FeatureSet::compute_score(const Context &context) {
  double score = 0;
  forvec(_, FeatureTemplate *, ft, templates)
    score += ft->score(context);
  return score;
}

// Compute the score of a tagged sequence.
double FeatureSet::compute_score(Sequence &seq, const IntVec &tags, bool use_averaged) {
  PtContext context(D, seq, use_averaged);
  double score = 0;
  forvec(i, int, tag, tags) {
    context.step(tag);
    assert((tag == D.boundary_token->tag) == (seq[i].word == D.boundary_token->word));
    score += seq[i].word == D.boundary_token->word ? 0 : compute_score(context);
  }
  return score;
}

// Compute the score of a segmented sequence.
double FeatureSet::compute_score(Sequence &seq, const SegVec &segs, bool use_averaged) {
  SegContext context(D, seq, use_averaged);
  double score = 0;
  forvec(i, Seg, seg, segs) {
    if(seg.a == -1) continue;
    context.step(seg);
    assert((seg == D.boundary_token->seg) == (seq[i].word == D.boundary_token->word));
    score += seq[i].word == D.boundary_token->word ? 0 : compute_score(context);
    // TMP: not compute features on O segments
    //score += seq[i].word == D.boundary_token->word || seg.a == D.no_seg ? 0 : compute_score(context);
  }
  return score;
}

void FeatureSet::print_feature_weights(const char *file, bool use_averaged) {
  track("print_feature_weights()", file, true);
  ofstream out(file); assert2(out, file);

  BigStatFig fig, nz_fig;
  forvec(_, FeatureTemplate *, ft, templates) {
    foridx(f, ft->params.size()) {
      double w = ft->params.get_weight(use_averaged, f);
      fig.add(w);
      if(feq(w, 0)) continue;

      nz_fig.add(w);
      out << ft->description(f) << '\t' << w << endl;
    }
  }

  logs("Non-zero feature weights: " << nz_fig);
  logs("All feature weights:      " << fig);
}

void FeatureSet::clear_plain_weights() {
  forvec(_, FeatureTemplate *, ft, templates) ft->params.clear_plain_weights();
}

void FeatureSet::begin_update() {
  forvec(_, FeatureTemplate *, ft, templates) ft->params.begin_update();
}

void FeatureSet::end_update() {
  forvec(_, FeatureTemplate *, ft, templates) ft->params.end_update();
}

void FeatureSet::dump_debug() {
  track("FeatureSet::dump_debug()", "", true);
  logs(prog_status);
  
  logs("Features");
  //log_info.out.add("f");
  int total = 0;
  forvec(_, FeatureTemplate *, ft, all_templates) {
    if(!ft->f_db && ft->params.size() == 0) continue;

    ostringstream os;
    os << ft->raw_description();
    if(ft->f_db) os << "\tf_db:" << ft->f_db << ",size=" << ft->f_db->size();
    os << ", " << ft->params.size() << " params";
    logs(os.str());
    total += ft->params.size();
  }
  logs(total << " TOTAL PARAMETERS");
  //log_info.out.remove_last();

  MemTracker mt;
  mt.add("word_db", D.word_db);
  mt.add("phrase_db", D.phrase_db);
  mt.add("params", total * (sizeof(real)*2+sizeof(int)));
  mt.report_mem_usage();

  logs(len(cached_templates) << " cached feature templates");

  MemTracker map_mt;
  forvec(_, TransformMap *, map, all_maps) {
    map_mt.add(map->transformer->description().c_str(), mem_usage(map->s2t)+mem_usage(map->t_db));
  }
  map_mt.report_mem_usage();
}

void FeatureSet::precompute_features(Sequence &seq) {
  // Precompute the feature values for each position and for each cached
  // feature template
  assert(::precompute_features);
  ::precompute_features = false;

  // Point
  PtContext pc(D, seq, false);
  for(pc.i = 0; pc.i < len(seq); pc.i++) {
    if(seq[pc.i].word == D.boundary_token->word) continue;
    forvec(_, FeatureTemplate *, ft, cached_templates) {
      if(ft->type() != FTYPE_PT) continue;
      IntDoubleVec feature_values;
      ft->get_feature_values(pc, feature_values);
    }
  }

  // Segment
  assert(max_seg_len != -1);
  SegContext sc(D, seq, false);
  for(sc.i = 0; sc.i < len(seq); sc.i++) {
    for(sc.n = 1; sc.n <= seq[sc.i].seg.n || (sc.n <= max_seg_len && sc.i+sc.n <= len(seq)); sc.n++) {
      if(seq[sc.i+sc.n-1].word == D.boundary_token->word) break;
      forvec(_, FeatureTemplate *, ft, cached_templates) {
        if(ft->type() != FTYPE_SEG) continue;
        IntDoubleVec feature_values;
        ft->get_feature_values(sc, feature_values);
      }
    }
  }
 
  ::precompute_features = true;
}

FeatureTemplate *FeatureSet::cache(FeatureTemplate *ft) {
  return push_back(cached_templates, mem(cache_ft(len(cached_templates), ft)));
}
