#ifndef __FEATURE_SET_H__
#define __FEATURE_SET_H__

#include "basic/stl-basic.h"
#include "basic/lisp.h"
#include "basic/str-str-db.h"
#include "domain.h"
#include "tag-features.h"
#include "seg-features.h"
#include "sequence.h"
#include "strength-db.h"

////////////////////////////////////////////////////////////

struct TransformMapSet {
  TransformMapSet();
  TransformMap *word_form_map, *brief_word_form_map, *word_lowercase_map;
  hash_map<string, TransformMap *, string_hf, string_eq> word_contains_map;
  hash_map<int, TransformMap *> word_prefix_map, word_suffix_map, word_cluster_path_map;
  hash_map<int, TransformMap *> crypt_maps;
};

class FeatureSet {
private:
  vector<FeatureTemplate *> templates; // The templates we care about
  vector<FeatureTemplate *> cached_templates;

  TransformMapSet word_maps, phrase_maps;

  vector<string> clusters_names;
  vector<StrStrDB> clusters_dbs; bool clusters_loaded;
  StrengthDB strength_db; bool strength_loaded;
  IntPairIntDB bigram_db;
  MapDB map_db; bool map_loaded;

  // To keep track of memory for deletion later.
  vector<FeatureTemplate *> all_templates;
  vector<TransformMap *> all_maps;
  vector<StringTransformer *> all_transformers;

  TransformMap *mem(TransformMap *map) { return push_back(all_maps, map); }
  FeatureTemplate *mem(FeatureTemplate *ft) { return push_back(all_templates, ft); }
  StringTransformer *mem(StringTransformer *transformer) { return push_back(all_transformers, transformer); }

public:
  int dp_order; // Order of the DP (1 = unigram, 2 = bigram, 3 = trigram)
  Domain &D;

public:
  FeatureSet(Domain &D);
  ~FeatureSet();

  void read_templates(const char *file);

  void print_tag_diff(Sequence &seq, const IntVec &truth_tags, const IntVec &pred_tags, bool use_averaged);
  void print_seg_diff(Sequence &seq, const SegVec &truth_segs, const SegVec &pred_segs, bool use_averaged);
  void print_feature_weights(const char *file, bool use_averaged);

  FeatureType type() const;
 
  void clear_weights();

  void update_weights_using_tags(Sequence &seq, const IntVec &tags, double incr);
  void update_weights_using_segs(Sequence &seq, const SegVec &segs, double incr);
  void update_weights(Sequence &seq, const IntVec &pred_tags);
  void update_weights(Sequence &seq, const SegVec &pred_segs);

  double compute_score(const Context &context);
  double compute_score(Sequence &seq, const IntVec &tags, bool use_averaged);
  double compute_score(Sequence &seq, const SegVec &segs, bool use_averaged);

  void clear_plain_weights();
  void begin_update();
  void end_update();

  void dump_debug();

  void precompute_features(Sequence &seq);

private:
  void parse_feature_templates(const LispNode *node, vector<FeatureTemplate *> &templates);
  void parse_map(const LispNode *node, vector<TransformMap *> &maps, StrDB &db, TransformMapSet &set);
  vector<TransformMap *> get_maps(const LispNode *node, StrDB &db, TransformMapSet &set);

  FeatureTemplate *cache(FeatureTemplate *ft);

  FeatureTemplate *lookup_feature_template(int f);
};

#endif
