#include "features.h"
#include "feature-cache.h"
#include "sequence.h"
#include "basic/str.h"
#include "basic/stl-utils.h"

extern bool precompute_features;

// f_db maps from a pair of feature indices to one feature index.
// But sometimes, we'd like to share f_db among feature templates.
// In fact, always sharing is correct, but sometimes not efficient (worst space
// space the parameters).  So, the product constructor takes a f_db;
// if it's NULL, then we use our own internal my_f_db.  This is a dangerous (hack):
// we MUST make sure that all feature templates are destructed at the very end.
#define define_product_constructor(t) \
  Product##t::Product##t(t &ft1, t &ft2, IntPairIntDB *f_db) : t(ft1.D), ft1(ft1), ft2(ft2) { \
    this->f_db = f_db ? f_db : &my_f_db; \
  }

#define define_fixed_product_constructor(t) \
  FixedProduct##t::FixedProduct##t(t &ft1, t &ft2) : t(ft1.D), ft1(ft1), ft2(ft2) { \
    assert(ft1.type() == ft2.type()); \
    assert(ft1.fixed_size != -1); \
    if(ft1.fixed_size == -1 || ft2.fixed_size == -1) \
      fixed_size = -1; \
    else \
      fixed_size = ft1.fixed_size * ft2.fixed_size; \
  }

#define define_cached_constructor(t) \
  Cached##t::Cached##t(int id, t &ft) : t(ft.D), ft(ft), id(id) { \
    assert(ft.only_on_input()); \
  }

////////////////////////////////////////////////////////////
// Multi

double MultiFeatureTemplate::score(const Context &context) {
  IntDoubleVec feature_values;
  get_feature_values(context, feature_values);
  double score = 0;
  forvec(_, const IntDouble &, fv, feature_values)
    score += params.get_weight(context.use_averaged, fv.first) * fv.second;
  return score;
}

void MultiFeatureTemplate::update_weights(const Context &context, double incr) {
  IntDoubleVec feature_values;
  get_feature_values(context, feature_values);
  forvec(_, const IntDouble &, fv, feature_values)
    params.update_weight(fv.first, fv.second * incr);
}

define_product_constructor(MultiFeatureTemplate)
define_fixed_product_constructor(MultiFeatureTemplate)
define_cached_constructor(MultiFeatureTemplate)

void ProductMultiFeatureTemplate::get_feature_values(const Context &context, IntDoubleVec &feature_values) {
  IntDoubleVec feature_values1, feature_values2;
  ft1.get_feature_values(context, feature_values1);
  ft2.get_feature_values(context, feature_values2);
  forvec(_, const IntDouble &, fv1, feature_values1)
    forvec(_, const IntDouble &, fv2, feature_values2)
      feature_values.push_back(IntDouble((*f_db)[IntPair(fv1.first, fv2.first)], fv1.second * fv2.second));
}

void FixedProductMultiFeatureTemplate::get_feature_values(const Context &context, IntDoubleVec &feature_values) {
  IntDoubleVec feature_values1, feature_values2;
  ft1.get_feature_values(context, feature_values1);
  ft2.get_feature_values(context, feature_values2);
  forvec(_, const IntDouble &, fv1, feature_values1)
    forvec(_, const IntDouble &, fv2, feature_values2)
      feature_values.push_back(IntDouble(fv1.first + ft1.fixed_size * fv2.first, fv1.second * fv2.second));
}

void CachedMultiFeatureTemplate::get_feature_values(const Context &context, IntDoubleVec &feature_values) {
  MultiFeatureCache &cache = context.seq.multi_cache(type());
  if(precompute_features) { cache.raw_get(context, id, feature_values); return; }
  if(!cache.get(context, id, feature_values)) {
    ft.get_feature_values(context, feature_values);
    cache.put(context, id, feature_values);
  }
}

////////////////////////////////////////////////////////////
// Single

void SingleFeatureTemplate::get_feature_values(const Context &context, IntDoubleVec &feature_values) {
  IntDouble fv = get_feature_value(context);
  feature_values.push_back(fv);
}

double SingleFeatureTemplate::score(const Context &context) {
  IntDouble fv = get_feature_value(context);
  return params.get_weight(context.use_averaged, fv.first) * fv.second;
}

void SingleFeatureTemplate::update_weights(const Context &context, double incr) {
  IntDouble fv = get_feature_value(context);
  params.update_weight(fv.first, fv.second * incr);
}

define_product_constructor(SingleFeatureTemplate)
define_fixed_product_constructor(SingleFeatureTemplate)
define_cached_constructor(SingleFeatureTemplate)

IntDouble ProductSingleFeatureTemplate::get_feature_value(const Context &context) {
  IntDouble fv1 = ft1.get_feature_value(context);
  IntDouble fv2 = ft2.get_feature_value(context);
  return IntDouble((*f_db)[IntPair(fv1.first, fv2.first)], fv1.second * fv2.second);
}

IntDouble FixedProductSingleFeatureTemplate::get_feature_value(const Context &context) {
  IntDouble fv1 = ft1.get_feature_value(context);
  IntDouble fv2 = ft2.get_feature_value(context);
  return IntDouble(fv2.first * ft1.fixed_size + fv1.first, fv1.second * fv2.second);
}

IntDouble CachedSingleFeatureTemplate::get_feature_value(const Context &context) {
  SingleFeatureCache &cache = context.seq.single_cache(type());
  if(precompute_features) return cache.raw_get(context, id);
  IntDouble fv = cache.get(context, id);
  if(fv.first == -1) {
    fv = ft.get_feature_value(context);
    cache.put(context, id, fv);
  }
  return fv;
}

////////////////////////////////////////////////////////////
// MultiBool

void MultiBoolFeatureTemplate::get_feature_values(const Context &context, IntDoubleVec &feature_values) {
  IntVec features;
  get_features(context, features);
  forvec(_, int, f, features)
    feature_values.push_back(IntDouble(f, 1));
}

double MultiBoolFeatureTemplate::score(const Context &context) {
  IntVec features;
  get_features(context, features);
  double score = 0;
  forvec(_, int, f, features)
    score += params.get_weight(context.use_averaged, f);
  return score;
}

void MultiBoolFeatureTemplate::update_weights(const Context &context, double incr) {
  IntVec features;
  get_features(context, features);
  forvec(_, int, f, features)
    params.update_weight(f, incr);
}

define_product_constructor(MultiBoolFeatureTemplate)
define_fixed_product_constructor(MultiBoolFeatureTemplate)
define_cached_constructor(MultiBoolFeatureTemplate)

void ProductMultiBoolFeatureTemplate::get_features(const Context &context, IntVec &features) {
  IntVec features1, features2;
  ft1.get_features(context, features1);
  ft2.get_features(context, features2);
  forvec(_, int, f1, features1)
    forvec(_, int, f2, features2)
      features.push_back((*f_db)[IntPair(f1, f2)]);
}

void FixedProductMultiBoolFeatureTemplate::get_features(const Context &context, IntVec &features) {
  IntVec features1, features2;
  ft1.get_features(context, features1);
  ft2.get_features(context, features2);
  forvec(_, int, f1, features1)
    forvec(_, int, f2, features2)
      features.push_back(f1 + ft1.fixed_size * f2);
}

void CachedMultiBoolFeatureTemplate::get_features(const Context &context, IntVec &features) {
  MultiBoolFeatureCache &cache = context.seq.multi_bool_cache(type());
  if(precompute_features) { cache.raw_get(context, id, features); return; }
  if(!cache.get(context, id, features)) {
    ft.get_features(context, features);
    cache.put(context, id, features);
  }
}

////////////////////////////////////////////////////////////
// SingleBool

void SingleBoolFeatureTemplate::get_features(const Context &context, IntVec &features) {
  int f = get_feature(context);
  features.push_back(f);
}

void SingleBoolFeatureTemplate::get_feature_values(const Context &context, IntDoubleVec &feature_values) {
  int f = get_feature(context);
  feature_values.push_back(IntDouble(f, 1));
}

double SingleBoolFeatureTemplate::score(const Context &context) {
  int f = get_feature(context);
  return params.get_weight(context.use_averaged, f);
}

void SingleBoolFeatureTemplate::update_weights(const Context &context, double incr) {
  int f = get_feature(context);
  params.update_weight(f, incr);
}

define_product_constructor(SingleBoolFeatureTemplate)
define_fixed_product_constructor(SingleBoolFeatureTemplate)
define_cached_constructor(SingleBoolFeatureTemplate)

int ProductSingleBoolFeatureTemplate::get_feature(const Context &context) {
  int f1 = ft1.get_feature(context);
  int f2 = ft2.get_feature(context);
  return (*f_db)[IntPair(f1, f2)];
}

int FixedProductSingleBoolFeatureTemplate::get_feature(const Context &context) {
  int f1 = ft1.get_feature(context);
  int f2 = ft2.get_feature(context);
  return f2 * ft1.fixed_size + f1;
}

int CachedSingleBoolFeatureTemplate::get_feature(const Context &context) {
  SingleBoolFeatureCache &cache = context.seq.single_bool_cache(type());
  if(precompute_features) return cache.raw_get(context, id);
  int f = cache.get(context, id);
  if(f == -1) {
    f = ft.get_feature(context);
    cache.put(context, id, f);
  }
  return f;
}

////////////////////////////////////////////////////////////

ThresholdSingleBoolFeatureTemplate::ThresholdSingleBoolFeatureTemplate(SingleFeatureTemplate &ft, double threshold)
  : SingleBoolFeatureTemplate(ft.D), ft(ft), threshold(threshold) {
  if(ft.fixed_size != -1) fixed_size = ft.fixed_size * 2;
}

// Each feature of ft expands into two indicator features; for below threshold and one for above threshold
int ThresholdSingleBoolFeatureTemplate::get_feature(const Context &context) {
  IntDouble fv = ft.get_feature_value(context);
  return fv.first * 2 + (fv.second >= threshold);
}

string ThresholdSingleBoolFeatureTemplate::description() const {
  return str_printf("%s>=%f?", ft.description().c_str(), threshold);
}

string ThresholdSingleBoolFeatureTemplate::description(int f) const {
  return str_printf("%s%s%f", ft.description(f/2).c_str(), f%2?">=":"<", threshold);
}

string ThresholdSingleBoolFeatureTemplate::raw_description() const {
  return "ThresholdSingleBoolFeatureTemplate("+ft.raw_description()+")";
}

////////////////////////////////////////////////////////////

FeatureTemplate *product_ft(FeatureTemplate *ft1, FeatureTemplate *ft2, IntPairIntDB *f_db) {
  bool fixed_size = ft1->fixed_size != -1;
  bool is_single = ft1->is_single() && ft2->is_single();
  bool is_bool = ft1->is_bool() && ft2->is_bool();

  // Note: can't cast SingleBool to Single, because SingleBool dervives from MultiBool,
  // because I don't know how to elegantly handle multiple inheritance.
  if(ft1->is_bool() ^ ft2->is_bool()) { // Exactly one is bool
    is_single = false;
  }

  if(fixed_size) {
    if(is_single) {
      if(is_bool) return new FixedProductSingleBoolFeatureTemplate(*(SingleBoolFeatureTemplate *)ft1, *(SingleBoolFeatureTemplate *)ft2);
      return new FixedProductSingleFeatureTemplate(*(SingleFeatureTemplate *)ft1, *(SingleFeatureTemplate *)ft2);
    }
    if(is_bool) return new FixedProductMultiBoolFeatureTemplate(*(MultiBoolFeatureTemplate *)ft1, *(MultiBoolFeatureTemplate *)ft2);
    return new FixedProductMultiFeatureTemplate(*(MultiFeatureTemplate *)ft1, *(MultiFeatureTemplate *)ft2);
  }
  if(is_single) {
    if(is_bool) return new ProductSingleBoolFeatureTemplate(*(SingleBoolFeatureTemplate *)ft1, *(SingleBoolFeatureTemplate *)ft2, f_db);
    return new ProductSingleFeatureTemplate(*(SingleFeatureTemplate *)ft1, *(SingleFeatureTemplate *)ft2, f_db);
  }
  if(is_bool) return new ProductMultiBoolFeatureTemplate(*(MultiBoolFeatureTemplate *)ft1, *(MultiBoolFeatureTemplate *)ft2, f_db);
  return new ProductMultiFeatureTemplate(*(MultiFeatureTemplate *)ft1, *(MultiFeatureTemplate *)ft2, f_db);
}

FeatureTemplate *cache_ft(int id, FeatureTemplate *ft) {
  bool is_single = ft->is_single();
  bool is_bool = ft->is_bool();

  if(is_single) {
    if(is_bool) return new CachedSingleBoolFeatureTemplate(id, *(SingleBoolFeatureTemplate *)ft);
    return new CachedSingleFeatureTemplate(id, *(SingleFeatureTemplate *)ft);
  }
  if(is_bool) return new CachedMultiBoolFeatureTemplate(id, *(MultiBoolFeatureTemplate *)ft);
  return new CachedMultiFeatureTemplate(id, *(MultiFeatureTemplate *)ft);
}
