#ifndef __FEATURES_H__
#define __FEATURES_H__

#include "domain.h"
#include "context.h"
#include "params.h"
#include "str-transform.h"
#include <string>

using namespace std;

// Inheritance tree:
// MultiFeatureTemplate ->
//   SingleFeatureTemplate
//   MultiBoolFeatureTemplate ->
//     SingleBoolFeatureTemplate 

#define define_general_product(t, s, b, f1, f2) \
  string description()      const { return ft1.description() + "," + ft2.description(); } \
  string description(int f) const { return ft1.description(f1) + "," + ft2.description(f2); } \
  string raw_description()  const { return __STRING(t)"("+ft1.raw_description()+","+ft2.raw_description()+")"; } \
  bool is_single()          const { return s; } \
  bool is_bool()            const { return b; } \
  bool only_on_input()      const { return ft1.only_on_input() && ft2.only_on_input(); } \
  FeatureType type()        const { return ft1.type(); }
#define define_product(t, s, b) \
  define_general_product(Product##t, s, b, (*f_db)[f].first, (*f_db)[f].second) \
  t &ft1, &ft2;
#define define_fixed_product(t, s, b) \
  define_general_product(FixedProduct##t, s, b, f % ft1.fixed_size, f / ft1.fixed_size) \
  t &ft1, &ft2;

#define define_cache(t, s, b) \
  string description()      const { return ft.description(); } \
  string description(int f) const { return ft.description(f); } \
  string raw_description()  const { return __STRING(Cached##t)"("+ft.raw_description()+")"; } \
  bool is_single()          const { return s; } \
  bool is_bool()            const { return b; } \
  bool only_on_input()      const { return true; } \
  FeatureType type()        const { return ft.type(); } \
  t &ft; \
  int id;

////////////////////////////////////////////////////////////

// Score = feature value * weight.
// Only score() and update_weights() need to be efficient.
class FeatureTemplate {
public:
  FeatureTemplate(Domain &D) : D(D), fixed_size(-1), f_db(NULL) { }
  virtual ~FeatureTemplate() { }

  virtual void get_feature_values(const Context &context, IntDoubleVec &feature_values) = 0;

  virtual double score(const Context &context) = 0;
  virtual void update_weights(const Context &context, double incr) = 0;

  virtual string description()      const = 0;
  virtual string description(int f) const = 0;
  virtual string raw_description()  const { return description(); }

  virtual bool is_single()     const = 0; // Only one feature is non-zero?
  virtual bool is_bool()       const = 0; // Feature values are boolean?
  virtual bool only_on_input() const = 0; // Features defined only on input
  virtual FeatureType type()   const = 0; // Feature over point or segment?

  Domain &D;
  Params params;
  int fixed_size; // Number of features if it doesn't change after initialization; -1 if it does.
  IntPairIntDB my_f_db, *f_db; // Used only for product feature templates
};

////////////////////////////////////////////////////////////
// We have two types of meta-FeatureTemplates: product and cache.
// For each of the two, there are many variants for efficient.
// You get the generality of having all, but pay only as much
// as you need.  Unfortunately, there's kind of an exponential
// blowup in how many types.  We can vary the following
//   - fixed_size (for product)
//   - single
//   - bool
//   - type of feature (pt or segment)

////////////////////////////////////////////////////////////
// Multi: single = false, bool = false

// For any context, exactly one feature from this feature template is non-zero.
class MultiFeatureTemplate : public FeatureTemplate {
public:
  MultiFeatureTemplate(Domain &D) : FeatureTemplate(D) { }
  double score(const Context &context);
  void update_weights(const Context &context, double incr);
  virtual bool is_single() const { return false; }
  virtual bool is_bool()   const { return false; }
};

struct ProductMultiFeatureTemplate : public MultiFeatureTemplate {
  ProductMultiFeatureTemplate(MultiFeatureTemplate &ft1, MultiFeatureTemplate &ft2, IntPairIntDB *f_db);
  void get_feature_values(const Context &context, IntDoubleVec &feature_values);
  define_product(MultiFeatureTemplate, false, false);
};

struct FixedProductMultiFeatureTemplate : public MultiFeatureTemplate {
  FixedProductMultiFeatureTemplate(MultiFeatureTemplate &ft1, MultiFeatureTemplate &ft2);
  void get_feature_values(const Context &context, IntDoubleVec &feature_values);
  define_fixed_product(MultiFeatureTemplate, false, false);
};

struct CachedMultiFeatureTemplate : public MultiFeatureTemplate {
  CachedMultiFeatureTemplate(int id, MultiFeatureTemplate &ft);
  void get_feature_values(const Context &context, IntDoubleVec &feature_values);
  define_cache(MultiFeatureTemplate, false, false);
};

////////////////////////////////////////////////////////////
// Single: single = true, bool = false

// For any context, exactly one feature from this feature template is non-zero.
class SingleFeatureTemplate : public MultiFeatureTemplate {
public:
  SingleFeatureTemplate(Domain &D) : MultiFeatureTemplate(D) { }
  void get_feature_values(const Context &context, IntDoubleVec &feature_values);
  double score(const Context &context);
  void update_weights(const Context &context, double incr);
  virtual bool is_single() const { return true; }
  virtual bool is_bool()   const { return false; }

  virtual IntDouble get_feature_value(const Context &context) = 0;
};

struct ProductSingleFeatureTemplate : public SingleFeatureTemplate {
  ProductSingleFeatureTemplate(SingleFeatureTemplate &ft1, SingleFeatureTemplate &ft2, IntPairIntDB *f_db);
  IntDouble get_feature_value(const Context &context);
  define_product(SingleFeatureTemplate, true, false);
};

struct FixedProductSingleFeatureTemplate : public SingleFeatureTemplate {
  FixedProductSingleFeatureTemplate(SingleFeatureTemplate &ft1, SingleFeatureTemplate &ft2);
  IntDouble get_feature_value(const Context &context);
  define_fixed_product(SingleFeatureTemplate, true, false);
};

struct CachedSingleFeatureTemplate : public SingleFeatureTemplate {
  CachedSingleFeatureTemplate(int id, SingleFeatureTemplate &ft);
  IntDouble get_feature_value(const Context &context);
  define_cache(SingleFeatureTemplate, true, false);
};

////////////////////////////////////////////////////////////
// MultiBool: single = false, bool = true

// Feature values are boolean.
class MultiBoolFeatureTemplate : public MultiFeatureTemplate {
public:
  MultiBoolFeatureTemplate(Domain &D) : MultiFeatureTemplate(D) { }
  void get_feature_values(const Context &context, IntDoubleVec &feature_values);
  double score(const Context &context);
  void update_weights(const Context &context, double incr);
  virtual bool is_single() const { return false; }
  virtual bool is_bool()   const { return true; }

  virtual void get_features(const Context &context, IntVec &features) = 0;
};

struct ProductMultiBoolFeatureTemplate : public MultiBoolFeatureTemplate {
  ProductMultiBoolFeatureTemplate(MultiBoolFeatureTemplate &ft1, MultiBoolFeatureTemplate &ft2, IntPairIntDB *f_db);
  void get_features(const Context &context, IntVec &features);
  define_product(MultiBoolFeatureTemplate, false, true);
};

struct FixedProductMultiBoolFeatureTemplate : public MultiBoolFeatureTemplate {
  FixedProductMultiBoolFeatureTemplate(MultiBoolFeatureTemplate &ft1, MultiBoolFeatureTemplate &ft2);
  void get_features(const Context &context, IntVec &features);
  define_fixed_product(MultiBoolFeatureTemplate, false, true);
};

struct CachedMultiBoolFeatureTemplate : public MultiBoolFeatureTemplate {
  CachedMultiBoolFeatureTemplate(int id, MultiBoolFeatureTemplate &ft);
  void get_features(const Context &context, IntVec &features);
  define_cache(MultiBoolFeatureTemplate, false, true);
};

////////////////////////////////////////////////////////////
// SingleBool: single = true, bool = true

// Exactly one feature has value 1, the rest have value 0.
class SingleBoolFeatureTemplate : public MultiBoolFeatureTemplate {
public:
  SingleBoolFeatureTemplate(Domain &D) : MultiBoolFeatureTemplate(D) { }
  void get_features(const Context &context, IntVec &features);
  void get_feature_values(const Context &context, IntDoubleVec &feature_values);
  double score(const Context &context);
  void update_weights(const Context &context, double incr);
  virtual bool is_single() const { return true; }
  virtual bool is_bool()   const { return true; }

  virtual int get_feature(const Context &context) = 0;
};

struct ProductSingleBoolFeatureTemplate : public SingleBoolFeatureTemplate {
  ProductSingleBoolFeatureTemplate(SingleBoolFeatureTemplate &ft1, SingleBoolFeatureTemplate &ft2, IntPairIntDB *f_db);
  int get_feature(const Context &context);
  define_product(SingleBoolFeatureTemplate, true, true);
};

struct FixedProductSingleBoolFeatureTemplate : public SingleBoolFeatureTemplate {
  FixedProductSingleBoolFeatureTemplate(SingleBoolFeatureTemplate &ft1, SingleBoolFeatureTemplate &ft2);
  int get_feature(const Context &context);
  define_fixed_product(SingleBoolFeatureTemplate, true, true);
};

struct CachedSingleBoolFeatureTemplate : public SingleBoolFeatureTemplate {
  CachedSingleBoolFeatureTemplate(int id, SingleBoolFeatureTemplate &ft);
  int get_feature(const Context &context);
  define_cache(SingleBoolFeatureTemplate, true, true);
};

////////////////////////////////////////////////////////////

// Converts a real number to 1 or 0
struct ThresholdSingleBoolFeatureTemplate : public SingleBoolFeatureTemplate {
  ThresholdSingleBoolFeatureTemplate(SingleFeatureTemplate &ft, double threshold);
  int get_feature(const Context &context);
  string description() const;
  string description(int f) const;
  string raw_description() const;
  bool only_on_input() const { return ft.only_on_input(); }
  FeatureType type() const { return ft.type(); }

  SingleFeatureTemplate &ft;
  double threshold;
};

////////////////////////////////////////////////////////////

FeatureTemplate *product_ft(FeatureTemplate *ft1, FeatureTemplate *ft2, IntPairIntDB *f_db);
FeatureTemplate *cache_ft(int id, FeatureTemplate *ft);

#undef define_general_product
#undef define_product
#undef define_fixed_product
#undef define_cache

#endif
