#include "params.h"
#include "basic/stl-utils.h"
#include "basic/logging.h"

extern bool use_winnow;
extern double winnow_rate;

////////////////////////////////////////////////////////////

void Params::resize(int nf) {
  if(nf > this->nf) {
    weights.resize(nf);
    averaged_weights.resize(nf);
    last_updated_time.resize(nf);
#ifdef WINNOW_CAPABLE
    if(use_winnow) {
      // In winnow, weights are initialized to 1.
      for(int f = this->nf; f < nf; f++) weights[f] = 1;
    }
#endif
    this->nf = nf;
  }
}

double Params::get_weight(bool use_averaged, int f) {
  resize(f+1);
  if(!use_averaged) return weights[f];

  update_weight(f, 0); // Refresh weight.
  return averaged_weights[f];
}

// File format: <feature_1> <weight_1> ...
void Params::read(const char *file) {
  ifstream in(file);
  assert(in);

  int nf;
  in >> nf;
  resize(nf);
  clear_weights();
  int f;
  double w, aw;
  while(in >> f >> w >> aw) {
    assert(f >= 0 && f < nf);
    weights[f] = w;
    averaged_weights[f] = aw;
  }
}

void Params::write(const char *file) {
  ofstream out(file);
  assert(out);
  int nf = len(weights);
  out << nf << endl;
  foridx(f, nf) {
    double w = weights[f], aw = averaged_weights[f];
    if(feq(w, 0) && feq(aw, 0)) continue;
    out << f << ' ' << w << ' ' << aw << endl;
  }
}

int Params::num_nonzero_weights() const {
  int n = 0;
  forvec(_, double, w, weights) {
    if(!feq(w, 0)) n++;
  }
  return n;
}

bool Params::should_keep_feature(int f) const {
  return len(keep_feature) == 0 || keep_feature[f];
}

void Params::update_weight(int f, double incr) {
  resize(f+1);

#ifdef WINNOW_CAPABLE
  if(use_winnow) {
    double factor = exp(winnow_rate*incr);
    //averaged_weights[f] += (curr_time - last_updated_time[f] - 1) * weights[f] + weights[f]*factor;
    last_updated_time[f] = curr_time;
    //logs(f << ' ' << weights[f] << ' ' << factor);
    weights[f] *= factor;
    averaged_weights[f] = weights[f];
  }
  else
#endif
  {
    averaged_weights[f] += (curr_time - last_updated_time[f]) * weights[f] + incr;
    last_updated_time[f] = curr_time;
    weights[f] += incr;
  }
}

int Params::set_keep_feature_from_weights() {
  keep_feature.resize(len(weights));
  int n = 0;
  foridx(f, nf) {
    if(weights[f] > 0)
      keep_feature[f] = 1, n++;
  }
  return n;
}

void Params::clear_plain_weights() {
  foridx(f, nf) weights[f] = 0;
}

void Params::clear_weights() {
  foridx(f, nf) {
    weights[f] = averaged_weights[f] = 0;
    last_updated_time[f] = 0;
  }
  curr_time = 0;
}

DoublePair Params::checksum() const {
  double sum = 0, asum = 0;
  foridx(f, nf) {
    sum += fabs(weights[f]);
    asum += fabs(averaged_weights[f]);
  }
  return DoublePair(sum, asum);
}
