#include "performance.h"
#include "sequence.h"

void Performance::update_stats(const Sequence &seq, const IntVec &truth_tags, const IntVec &pred_tags) {
  update_stats_tag(seq, truth_tags, pred_tags);
  if(D.is_seg())
    update_stats_seg(seq, Sequence::tags_to_segs(D, truth_tags), Sequence::tags_to_segs(D, pred_tags));
}

void Performance::update_stats(const Sequence &seq, const SegVec &truth_segs, const SegVec &pred_segs) {
  update_stats_tag(seq, Sequence::segs_to_tags(D, truth_segs), Sequence::segs_to_tags(D, pred_segs));
  update_stats_seg(seq, truth_segs, pred_segs);
}

// Evaluate tag performance.
void Performance::update_stats_tag(const Sequence &seq, const IntVec &truth_tags, const IntVec &pred_tags) {
  int n = D.tag_db.size();
  matrix_resize(tag_confusion, n, n);

  bool all_correct = true;
  foridx(i, len(truth_tags)) {
    if(seq[i].word == D.boundary_token->word) continue;

    bool correct = truth_tags[i] == pred_tags[i];
    all_correct &= correct;
    word_error.add(!correct);
    tag_confusion[truth_tags[i]][pred_tags[i]]++;
  }

  sentence_error.add(!all_correct);
}

struct show_Segment {
  show_Segment(const Domain &D, const Sequence &seq, int a, int i, int n)
    : D(D), seq(seq), a(a), i(i), n(n) { }
  const Domain &D;
  const Sequence &seq;
  int a, i, n;
};
ostream &operator<<(ostream &out, const show_Segment &show) {
  if(show.a == -1 || show.a == show.D.no_seg) return out << "NONE";
  out << show.D.word_db[show.seq[show.i-1].word] << ' ' << show.D.seg_db[show.a] << "(";
  for(int k = 0; k < show.n; k++) {
    if(k > 0) out << ' ';
    out << show.D.word_db[show.seq[show.i+k].word];
  }
  out << ") " << show.D.word_db[show.seq[show.i+show.n].word];
  return out;
}

void Performance::update_stats_seg(const Sequence &seq, const SegVec &truth_segs, const SegVec &pred_segs) {
  track("update_stats_seg()", "", true);
  assert(D.is_seg());
  int n = D.seg_db.size();
  matrix_resize(seg_confusion, n, n);

  foridx(i, len(truth_segs)) {
    if(seq[i].word == D.boundary_token->word) continue;

    Seg truth_seg = truth_segs[i];
    Seg pred_seg = pred_segs[i];

    int truth_seg_a = truth_seg.a == -1 ? D.no_seg : truth_seg.a; // Let -1 be equivalent to D.no_seg
    int pred_seg_a = pred_seg.a == -1 ? D.no_seg : pred_seg.a;

    bool correct = truth_seg_a == pred_seg_a && truth_seg.n == pred_seg.n;
    /*if(!correct) {
      logs("Truth seg at i=" << i << ": " << show_Segment(D, seq, truth_seg_a, i, truth_seg.n));
      logs("Pred  seg at i=" << i << ": " << show_Segment(D, seq, pred_seg_a, i, pred_seg.n));
    }*/

    // Don't count no_seg in the precision and recall estimates.
    // The confusion matrix is kind of strange because it presupposes some match between the
    // truth segments and the pred segments.

    if(truth_seg_a != D.no_seg) {
      seg_confusion[truth_seg_a][pred_seg_a != D.no_seg && truth_seg.n == pred_seg.n ? pred_seg_a : D.no_seg]++;
      recall.add(correct);
    }
    if(pred_seg_a != D.no_seg) {
      seg_confusion[truth_seg_a != D.no_seg && pred_seg.n == truth_seg.n ? truth_seg_a : D.no_seg][pred_seg_a]++;
      precision.add(correct);
    }
    if(truth_seg_a != D.no_seg && pred_seg_a != D.no_seg && correct) // Don't double count.
      seg_confusion[truth_seg_a][pred_seg_a]--;
  }
}

void Performance::print_confusion() const {
  if(D.is_seg())
    print_confusion(D.seg_db, seg_confusion, D.boundary_token->tag);
  print_confusion(D.tag_db, tag_confusion, D.boundary_token->tag);
}

void Performance::print_confusion(const StrDB &db, const IntVecVec &confusion, int skip) const {
#define show(a) ((a) != skip)
  // Print confusion matrix.
  int n = len(confusion);
  const int W = 10;

  ostringstream sout;
  sout << setw(W) << "Truth\\Pred";
  foridx(b, n)
    if(show(b))
      sout << setw(W) << db[b];
  logss(sout.str());

  foridx(a, n) {
    if(!show(a)) continue;
    ostringstream sout;
    sout << setw(W) << db[a];
    foridx(b, n)
      if(show(b))
        sout << setw(W) << confusion[a][b];
    logss(sout.str());
  }
#undef show
}

void Performance::operator=(const Performance &perf) {
  sentence_error = perf.sentence_error;
  word_error = perf.word_error;
  precision = perf.precision;
  recall = perf.recall;
  tag_confusion = perf.tag_confusion;
  seg_confusion = perf.seg_confusion;
  errors = perf.errors;
}

ostream &operator<<(ostream &out, const Performance &perf) {
  out << "sentence error: " << perf.sentence_error
      << ", word error: " << perf.word_error;
  if(perf.D.is_seg())
    out << ", precision: " << perf.precision
        << ", recall: " << perf.recall
        << ", F1: " << perf.F1();
  return out;
}
