#include "seg-features.h"
#include "features-macros.h"
#include "basic/str.h"
#include "sequence.h"

//////////////////////////////
// prev_seg, curr_seg

F_PrevSeg_CurrSeg::F_PrevSeg_CurrSeg(Domain &D)
  : SingleBoolFeatureTemplate(D) {
  fixed_size = nsegs * nsegs;
}

int F_PrevSeg_CurrSeg::get_feature(const Context &context) {
  SegContext &c = (SegContext &)context;
  return encode2(c.prev_seg, nsegs, c.curr_seg, nsegs);
}

string F_PrevSeg_CurrSeg::description() const { return "prev_seg,curr_seg"; }

string F_PrevSeg_CurrSeg::description(int f) const {
  int prev_seg, curr_seg;
  decode2(f, prev_seg, nsegs, curr_seg, nsegs);
  return str_printf("prev_seg=%s,curr_seg=%s", D.seg_db[prev_seg], D.seg_db[curr_seg]);
}

//////////////////////////////
// curr_seg

F_CurrSeg::F_CurrSeg(Domain &D) : SingleBoolFeatureTemplate(D) {
  fixed_size = nsegs;
}

int F_CurrSeg::get_feature(const Context &context) {
  SegContext &c = (SegContext &)context;
  return encode1(c.curr_seg, nsegs);
}

string F_CurrSeg::description() const { return "curr_seg"; }

string F_CurrSeg::description(int f) const {
  int curr_seg;
  decode1(f, curr_seg, nsegs);
  return str_printf("curr_seg=%s", D.seg_db[curr_seg]);
}

//////////////////////////////
// phrase

F_Phrase::F_Phrase(Domain &D, TransformMap *map)
  : SingleBoolFeatureTemplate(D), map(map) { }

int F_Phrase::get_feature(const Context &context) {
  SegContext &c = (SegContext &)context;
  int p = D.phrase_db[c.seq.phrase(D, c.i, c.n).c_str()];
  assert(!map || &map->s_db == &D.phrase_db); // Better be the same db
  return map ? (*map)[p] : p;
}

string F_Phrase::description() const {
  return map ? str_printf("%s(phrase)", map->description().c_str()) :
               str_printf("phrase");
}

string F_Phrase::description(int f) const {
  return map ? str_printf("%s(phrase)=%s", map->description().c_str(), map->str(f)) :
               str_printf("phrase=%s", D.phrase_db[f]);
}

//////////////////////////////
// POS phrase

F_POSPhrase::F_POSPhrase(Domain &D) : SingleBoolFeatureTemplate(D) { }

int F_POSPhrase::get_feature(const Context &context) {
  SegContext &c = (SegContext &)context;
  // Overload phrase_db to store not just the text strings, but strings of POS
  return D.phrase_db[c.seq.pos_phrase(D, c.i, c.n).c_str()];
}

string F_POSPhrase::description() const {
  return str_printf("pos_phrase");
}

string F_POSPhrase::description(int f) const {
  return str_printf("pos_phrase=%s", D.phrase_db[f]);
}

//////////////////////////////
// focus

Focus parse_focus(const string &s) {
  if(s == "BEGIN") return FOCUS_BEGIN;
  if(s == "END") return FOCUS_END;
  assert2(false, "Expected either BEGIN or END, got " << s);
}

string focus_str(Focus focus) {
  if(focus == FOCUS_BEGIN) return "begin";
  if(focus == FOCUS_END) return "end";
  assert(0);
}

F_SingleBoolFocusPhrase::F_SingleBoolFocusPhrase(SingleBoolFeatureTemplate &ft, Focus focus) :
  SingleBoolFeatureTemplate(ft.D), ft(ft), focus(focus) {
  assert(ft.type() == FTYPE_PT);
  assert(ft.only_on_input());
  fixed_size = ft.fixed_size; 
}

int F_SingleBoolFocusPhrase::get_feature(const Context &context) {
  SegContext &c = (SegContext &)context;
  PtContext pc(c.D, c.seq, c.use_averaged);
  pc.i = (focus == FOCUS_BEGIN ? c.i : c.i+c.n-1);
  return ft.get_feature(pc);
}

string F_SingleBoolFocusPhrase::raw_description() const {
  return "F_SingleBoolFocusPhrase("+focus_str(focus)+","+ft.raw_description()+")";
}

string F_SingleBoolFocusPhrase::description() const {
  return str_printf("focus_%s:%s",
      focus_str(focus).c_str(),
      ft.description().c_str());
}

string F_SingleBoolFocusPhrase::description(int f) const {
  return str_printf("focus_%s:%s",
      focus_str(focus).c_str(),
      ft.description(f).c_str());
}

F_FocusPhrase::F_FocusPhrase(MultiFeatureTemplate &ft, Focus focus) :
  MultiFeatureTemplate(ft.D), ft(ft), focus(focus) {
  assert(ft.type() == FTYPE_PT);
  assert(ft.only_on_input());
  fixed_size = ft.fixed_size; 
}

void F_FocusPhrase::get_feature_values(const Context &context, IntDoubleVec &feature_values) {
  SegContext &c = (SegContext &)context;
  PtContext pc(c.D, c.seq, c.use_averaged);
  pc.i = (focus == FOCUS_BEGIN ? c.i : c.i+c.n-1);
  ft.get_feature_values(pc, feature_values);
}

string F_FocusPhrase::raw_description() const {
  return "F_FocusPhrase("+focus_str(focus)+","+ft.raw_description()+")";
}

string F_FocusPhrase::description() const {
  return str_printf("focus_%s:%s",
      focus_str(focus).c_str(),
      ft.description().c_str());
}

string F_FocusPhrase::description(int f) const {
  return str_printf("focus_%s:%s",
      focus_str(focus).c_str(),
      ft.description(f).c_str());
}

//////////////////////////////
// aggregate

F_AggregatePhrase::F_AggregatePhrase(MultiFeatureTemplate &ft, int start) :
    MultiFeatureTemplate(ft.D), ft(ft), start(start) {
  assert(ft.type() == FTYPE_PT);
  assert(ft.only_on_input());
  fixed_size = ft.fixed_size;
}

void F_AggregatePhrase::get_feature_values(const Context &context, IntDoubleVec &feature_values) {
  SegContext &c = (SegContext &)context;
  PtContext pc(c.D, c.seq, c.use_averaged);
  for(pc.i = c.i+start; pc.i < c.i+c.n; pc.i++)
    ft.get_feature_values(pc, feature_values);
}

string F_AggregatePhrase::raw_description() const {
  return str_printf("F_AggregatePhrase(%+d,%s)", start, ft.raw_description().c_str());
}

string F_AggregatePhrase::description() const {
  return str_printf("aggregate%+d:%s", start, ft.description().c_str());
}

string F_AggregatePhrase::description(int f) const {
  return str_printf("aggregate%+d:%s", start, ft.description(f).c_str());
}

//////////////////////////////
// length

F_PhraseLength::F_PhraseLength(Domain &D)
  : SingleBoolFeatureTemplate(D) { }

int F_PhraseLength::get_feature(const Context &context) {
  SegContext &c = (SegContext &)context;
  return c.n;
}

string F_PhraseLength::description() const {
  return str_printf("length(phrase)");
}

string F_PhraseLength::description(int f) const {
  int n = f;
  return str_printf("length(phrase)=%d", n);
}

//////////////////////////////
// strength

F_PhraseStrength::F_PhraseStrength(Domain &D, const StrengthDB &strength_db, double scale, double power, int idx)
  : SingleFeatureTemplate(D), strength_db(strength_db), scale(scale), power(power), idx(idx) {
  fixed_size = 1;
  assert(idx >= 0 && idx < strength_db.nstrengths);
}

IntDouble F_PhraseStrength::get_feature_value(const Context &context) {
  SegContext &c = (SegContext &)context;
  IntVec seq;
  for(int i = c.i; i < c.i+c.n; i++)
    seq.push_back(c.seq[i].word);
  double strength = strength_db.get_strength(seq, idx);

  // Only has one feature (whose value is the scale*strength^power)
  return IntDouble(0, scale*pow(strength, power));
}

string F_PhraseStrength::description() const {
  return str_printf("%.1f*strength(%d..%d)[%d]^%.1f", scale, -1, 0, idx, power);
}

string F_PhraseStrength::description(int f) const {
  assert(f == 0);
  return description();
}
