#ifndef __STR_TRANSFORM_H__
#define __STR_TRANSFORM_H__

#include "basic/strdb.h"
#include "basic/str-str-db.h"

////////////////////////////////////////////////////////////

class StringTransformer {
public:
  virtual ~StringTransformer() { }
  virtual string transform(const char *s) const = 0;
  virtual string description() const = 0;
};

// t2(t1(s))
struct ComposeStringTransformer : StringTransformer {
  ComposeStringTransformer(const StringTransformer *t1, const StringTransformer *t2) : t1(t1), t2(t2) { }

  string transform(const char *s) const {
    return t2->transform(t1->transform(s).c_str());
  }
  virtual string description() const {
    return t2->description() + "\\" + t1->description();
  }

  const StringTransformer *t1, *t2;
};

////////////////////////////////////////////////////////////

// s -> s[0..k]
struct T_Prefix : public StringTransformer {
  T_Prefix(int k) : k(k) { }
  string transform(const char *s) const;
  string description() const;
  int k;
};

// s -> s[len(s)-k..len(s)]
struct T_Suffix : public StringTransformer {
  T_Suffix(int k) : k(k) { }
  string transform(const char *s) const;
  string description() const;
  int k;
};

// "TEST" -> "test"
struct T_Lowercase : public StringTransformer {
  string transform(const char *s) const;
  string description() const;
};

// "TEST" -> "Y" or "N" depending on whether "TEST" contains the appropriate description.
class T_Contains : public StringTransformer {
public:
  T_Contains(string descrip);
  string transform(const char *s) const;
  string description() const;
private:
  BoolVec hit;
  string descrip;
};

// s -> { "Aaaa", "Aa", "aa", "0", ... }
struct T_Form : public StringTransformer {
  T_Form(bool brief) : brief(brief) { }
  string transform(const char *s) const;
  string description() const;
  bool brief;
};

struct T_TableLookup : public StringTransformer {
  T_TableLookup(const char *name, const StrStrDB &db) : name(name), db(db) { }
  string transform(const char *s) const { return db[s]; }
  string description() const { return name; }
  string name;
  const StrStrDB &db;
};

struct TransformMap {
  TransformMap(const StringTransformer *transformer, StrDB &s_db) :
    transformer(transformer), s_db(s_db) { }

  int operator[](int s);
  int operator[](const char *s) { return (*this)[s_db[s]]; }
  const char *str(int t) const { return t_db[t]; }
  string description() const { return transformer->description(); }

  const StringTransformer *transformer;
  IntVec s2t;
  StrDB &s_db; // source strings
  StrDB t_db; // transformed strings
};

// s -> some crypt hash function of s (essentially a deterministic noise feature)
struct T_Crypt : public StringTransformer {
  T_Crypt(int key, int range) : key(key), range(range) { }
  string transform(const char *s) const;
  string description() const;
  int key, range;
};

#endif
