#include "tag-features.h"
#include "features-macros.h"
#include "basic/str.h"
#include "sequence.h"

//////////////////////////////
// prev2_tag, prev_tag, curr_tag

F_Prev2Tag_PrevTag_CurrTag::F_Prev2Tag_PrevTag_CurrTag(Domain &D)
  : SingleBoolFeatureTemplate(D) {
  fixed_size = ntags * ntags * ntags;
}

int F_Prev2Tag_PrevTag_CurrTag::get_feature(const Context &context) {
  PtContext &c = (PtContext &)context;
  return encode3(c.prev2_tag, ntags, c.prev_tag, ntags, c.curr_tag, ntags);
}

string F_Prev2Tag_PrevTag_CurrTag::description() const { return "prev2_tag,prev_tag,curr_tag"; }

string F_Prev2Tag_PrevTag_CurrTag::description(int f) const {
  int prev2_tag, prev_tag, curr_tag;
  decode3(f, prev2_tag, ntags, prev_tag, ntags, curr_tag, ntags);

  return str_printf("prev2_tag=%s,prev_tag=%s,curr_tag=%s",
      D.tag_db[prev2_tag], D.tag_db[prev_tag], D.tag_db[curr_tag]);
}

//////////////////////////////
// prev_tag, curr_tag

F_PrevTag_CurrTag::F_PrevTag_CurrTag(Domain &D)
  : SingleBoolFeatureTemplate(D) {
  fixed_size = ntags * ntags;
}

int F_PrevTag_CurrTag::get_feature(const Context &context) {
  PtContext &c = (PtContext &)context;
  return encode2(c.prev_tag, ntags, c.curr_tag, ntags);
}

string F_PrevTag_CurrTag::description() const { return "prev_tag,curr_tag"; }

string F_PrevTag_CurrTag::description(int f) const {
  int prev_tag, curr_tag;
  decode2(f, prev_tag, ntags, curr_tag, ntags);
  return str_printf("prev_tag=%s,curr_tag=%s", D.tag_db[prev_tag], D.tag_db[curr_tag]);
}

//////////////////////////////
// curr_tag

F_CurrTag::F_CurrTag(Domain &D) : SingleBoolFeatureTemplate(D) {
  fixed_size = ntags;
}

int F_CurrTag::get_feature(const Context &context) {
  PtContext &c = (PtContext &)context;
  return encode1(c.curr_tag, ntags);
}

string F_CurrTag::description() const { return "curr_tag"; }

string F_CurrTag::description(int f) const {
  int curr_tag;
  decode1(f, curr_tag, ntags);
  return str_printf("curr_tag=%s", D.tag_db[curr_tag]);
}

//////////////////////////////
// word(i+offset)

F_Word::F_Word(Domain &D, int offset, TransformMap *map)
  : SingleBoolFeatureTemplate(D), offset(offset), map(map) { }

int F_Word::get_feature(const Context &context) {
  PtContext &c = (PtContext &)context;
  int i = c.i + offset;
  return map ? (*map)[c.seq[i].word] : c.seq[i].word;
}

string F_Word::description() const {
  return map ? str_printf("%s(word(%d))", map->description().c_str(), offset) :
               str_printf("word(%d)", offset);
}

string F_Word::description(int f) const {
  return map ? str_printf("%s(word(%d))=%s", map->description().c_str(), offset, map->str(f)) :
               str_printf("word(%d)=%s", offset, D.word_db[f]);
}

//////////////////////////////
// words(i+offset)

F_Words::F_Words(Domain &D, int offset, int n, TransformMap *map)
  : SingleBoolFeatureTemplate(D), offset(offset), n(n), map(map) { }

int F_Words::get_feature(const Context &context) {
  PtContext &c = (PtContext &)context;
  int i = c.i + offset;
  int p = D.phrase_db[c.seq.phrase(D, i, n).c_str()];
  assert(!map || &map->s_db == &D.phrase_db); // Better be the same db
  return map ? (*map)[p] : p;
}

string F_Words::description() const {
  return map ? str_printf("%s(words(%d..%d))", map->description().c_str(), offset, offset+n-1) :
               str_printf("word(%d..%d)", offset, offset+n-1);
}

string F_Words::description(int f) const {
  return map ? str_printf("%s(word(%d..%d))=%s", map->description().c_str(), offset, offset+n-1, map->str(f)) :
               str_printf("word(%d..%d)=%s", offset, offset+n-1, D.phrase_db[f]);
}

//////////////////////////////
// pos(i+offset)

F_POS::F_POS(Domain &D, int offset) : SingleBoolFeatureTemplate(D), offset(offset) { }

int F_POS::get_feature(const Context &context) {
  PtContext &c = (PtContext &)context;
  int i = c.i + offset;
  return c.seq[i].pos;
}

string F_POS::description() const { return str_printf("pos(%d)", offset); }

string F_POS::description(int f) const {
  int pos = f;
  return str_printf("pos(%d)=%s", offset, D.pos_db[pos]);
}

//////////////////////////////
// strength

F_Strength::F_Strength(Domain &D, const StrengthDB &strength_db, double scale, double power, int idx)
  : SingleFeatureTemplate(D), strength_db(strength_db), scale(scale), power(power), idx(idx) {
  fixed_size = 1;
  assert(idx >= 0 && idx < strength_db.nstrengths);
}

IntDouble F_Strength::get_feature_value(const Context &context) {
  PtContext &c = (PtContext &)context;
  double strength = strength_db.get_strength(c.seq[c.i-1].word, c.seq[c.i].word, idx);
  /*if(strength > 0)
    logs("Strength(" << D.word_db[c.seq[c.i-1].word] << ' ' << D.word_db[c.seq[c.i].word] << ") = " << strength);*/

  // Only has one feature (whose value is the scale*strength^power)
  return IntDouble(0, scale*pow(strength, power));
}

string F_Strength::description() const {
  return str_printf("%.1f*strength(%d..%d)[%d]^%.1f", scale, -1, 0, idx, power);
}

string F_Strength::description(int f) const {
  assert(f == 0);
  return description();
}

//////////////////////////////
// map

F_Map::F_Map(Domain &D, int offset, int n, const MapDB &map_db) :
    MultiFeatureTemplate(D), offset(offset), n(n), map_db(map_db) {
  fixed_size = len(map_db);
}

void F_Map::get_feature_values(const Context &context, IntDoubleVec &feature_values) {
  PtContext &c = (PtContext &)context;

  int i = c.i + offset;
  map_db.get_mapping(c.seq.phrase(D, i, n).c_str(), feature_values);
}

string F_Map::description() const {
  return str_printf("map(%d..%d)", offset, offset+n-1);
}

string F_Map::description(int f) const {
  return str_printf("map(%d..%d)[%s]", offset, offset+n-1, map_db.get_name(f));
}
