/*
 * Tags a sequence.
 */

#include "basic/str.h"
#include "basic/opt.h"
#include "example-set.h"

opt_define_string(tag_domain_str, "tag-domain", "pos", "The type of tag we use (pos, chunk, ent).");
opt_define_string(prefix, "prefix", "", "Directory to store all the results in.");
opt_define_string(prepend_prefix, "prepend-prefix", "", "Add this to the beginning of the prefix name.");
opt_define_string(append_prefix, "append-prefix", "", "Add this to the end of the prefix name.");
opt_define_string(train_file, "train", "train.in", "File with training data (input).");
opt_define_string(test_file, "test", "dev.in", "File with test data (input).");
opt_define_string(test2_file, "test2", "", "File with test data (input).");
opt_define_string(features_file, "features", "features.in", "File containing the features to use (input).");
opt_define_string(out_weights_file, "out-weights", "", "File to write feature weights.");
opt_define_string(clusters_file, "clusters", "", "File to read word clustering information (input).");
opt_define_string(strength_file, "strength", "", "File to read strengths about grouping of words.");
opt_define_string(map_file, "map", "", "File to read the feature vector map from for MapDB.");
opt_define_string(read_params_file, "read-params", "", "Use weights from this file as an initial seed.");

opt_define_string(comm_file, "comm", "", "Write communication information to this file.");

opt_define_string(params_file, "params", "", "Feature parameters with weights (output).");
opt_define_string(stats_file, "stats", "", "Performance stats (output).");
opt_define_string(eval_file, "eval", "", "CoNLL 2003 evaluation format for named-entities (output).");

opt_define_int(niters, "niters", 10, "Number of iterations.");
opt_define_bool(eval_after_each_iter, "eval-after-each-iter", false, "Whether to evaluate the test set after each iteration as opposed to just the last iteration.");
opt_define_bool(print_features_after_update, "print-features-after-update", false, "Whether to print feature weights after each update.");
opt_define_bool(print_all_errors, "print-all-errors", false, "Whether to print all errors at end of testing.");
opt_define_bool(print_wrong_details, "print-wrong-details", false, "When we miss an example, print out details about why.");
opt_define_bool(debug, "debug", false, "Print everything.");
opt_define_bool(tmp, "tmp", false, "Temporary flag.");
opt_define_double(train_frac, "train-frac", 1, "Fraction of the training set to use.");
opt_define_int(train_limit, "train-limit", INT_MAX, "Limit number of examples.");
opt_define_int(test_limit, "test-limit", INT_MAX, "Limit number of test examples.");
opt_define_bool(use_beam, "use-beam", false, "Use beam search as opposed to the full DP.");
opt_define_int(beam_size, "beam-size", 20, "Size of the beam.");
opt_define_bool(beam_check, "beam-check", false, "Check the optimality of beam search.");
opt_define_int(nprefix_incr, "nprefix-incr", INT_MAX, "Use prefixes of length 0, n, 2n, 3n of the cluster path (0 corresponds to the whole string).");

opt_define_bool(lazy_run, "lazy-run", false, "Run only if there are no results.");
opt_define_bool(use_cache, "use-cache", true, "Use the feature cache.");
opt_define_bool(use_bpm, "use-bpm", false, "Randomize the order of the examples, and run Perceptron many times.");
opt_define_bool(test_only, "test-only", false, "Whether just to evaluate on the test data, not train.");
opt_define_bool(save_params, "save-params", true, "Whether to save the feature parameter weights to disk after training.");
opt_define_double(cross_validate_ratio, "cross-validate", 0.0, "Fraction of training examples to use for cross validation (set to 0 to disable).");
opt_define_int(max_seg_len, "max-seg-len", -1, "Maximum length of a segment that we will recover (when we are doing the DP on segments rather than BIO tags).");
//opt_define_int_req(max_seg_len, "max-seg-len", "Maximum length of a segment that we will recover (when we are doing the DP on segments rather than BIO tags).");
opt_define_int(boundary_size, "boundary-size", 3, "Padding around each sentence, so that accessing previous words doesn't result in out-of bounds");
opt_define_bool(precompute_features, "precompute-features", true, "Whether to precompute the features for all sequences before training.");

opt_define_double(winnow_rate, "winnow-rate", 0.1, "Learning rate for winnow.");
opt_define_bool(use_winnow, "use-winnow", false, "Use the winnow update rule instead of the Perceptron");

// Active learning.
opt_define_bool(use_active_learning, "active-learning", false, "Use active learning to choose examples.");
opt_define_bool(active_learning_sample_randomly, "al-random", false, "If using active learning, sample randomly.");
opt_define_bool(active_learning_sample_margin, "al-margin", false, "If using active learning, sample based on the inverse margin.");
opt_define_double(active_learning_incr, "al-incr", 0.2, "Number of examples to add each time actively; if <1, fraction of current");
opt_define_double(active_learning_ninit, "al-ninit", 100, "Number of examples to start with; if <1, fraction of total");
opt_define_bool(active_learning_norm_len, "al-norm-len", true, "Normalize the length of the sequence when measuring uncertainty");
opt_define_bool(active_learning_print_picked, "al-print-picked", false, "Print examples that are picked");

////////////////////////////////////////////////////////////

int main(int argc, char *argv[]) {
  init_opt(argc, argv);

  // Set prefix from arguments.
  if(prefix.empty()) {
    if(!prepend_prefix.empty()) prefix += append_prefix + "-";

    prefix += file_base(strip_dir(features_file));
    prefix += str_printf("-n%d", niters);
    prefix += "-f" + double2str(train_frac);
    if(use_bpm)                prefix += "-bpm";
    if(use_winnow)             prefix += "-winnow";
    if(!clusters_file.empty()) prefix += "-Fc_"+file_base(strip_dir(clusters_file));
    if(!strength_file.empty()) prefix += "-Fs_"+file_base(strip_dir(strength_file));
    if(!map_file.empty())      prefix += "-Fm_"+file_base(strip_dir(map_file));

    if(use_active_learning) {
      if(active_learning_sample_randomly)    prefix += "-AR";
      else if(active_learning_sample_margin) prefix += "-AM";
      else                                   prefix += "-AL";
      prefix += "-An" + double2str(active_learning_ninit) +
                  "+" + double2str(active_learning_incr);
    }

    if(!append_prefix.empty()) prefix += "-" + append_prefix;
    prefix += ".out";
    if(rand_seed != 42) prefix += str_printf("/%d", rand_seed);
  }

  // Used to pass messages to the program who called this
  // Write out where our directory is
  if(!comm_file.empty()) {
    ofstream out(comm_file.c_str());
    out << prefix << endl;
    out.close();
  }

  if(lazy_run) {
    ifstream in((prefix+"/status").c_str());
    char buf[256];
    // If the last process terminated properly, then we don't rerun it
    if(in && in.getline(buf, sizeof(buf) && !strncmp(buf, "term ", 5))) {
      cout << prefix << " either finished ok or still running; exiting...\n";
      return 0;
    }
  }

  if(system(("mkdir -p " + prefix).c_str()) != 0)
    assert2(false, "Can't create " << prefix);
  if(system(("rm -f " + prefix + "/*").c_str()) != 0)
    assert2(false, "Can't remove things in " << prefix);

  // Set arguments from the prefix.
  if(params_file.empty()) params_file = prefix + "/params";
  if(stats_file.empty()) stats_file = prefix + "/stats";
  if(eval_file.empty()) eval_file = prefix + "/eval";
  if(log_info.log_file.empty()) log_info.log_file = prefix+"/log";

  // Debugging mode.
  if(debug) {
    log_info.ms_per_line = 0;
    print_features_after_update = true;
    print_wrong_details = true;
  }
  if(test_only) {
    train_file = "/dev/null";
    niters = 1;
    save_params = false;
  }

  init_log;
  print_opts();

  // Input files must exist.
  assert(file_exists(train_file.c_str()));
  assert(file_exists(test_file.c_str()));
  assert(file_exists(features_file.c_str()));

  ExampleSet es(tag_domain_str);

  // 1) Read training and test sequences.
  es.read(train_file.c_str(), test_file.c_str(), test2_file.c_str(), train_frac);
  if(cross_validate_ratio > 0)
    es.steal_train_data_for_cross_validation(cross_validate_ratio);
  es.finish_initialization();

  // 3) Read feature templates (must follow reading in the sequences).
  es.features.read_templates(features_file.c_str());

  // 4) Train and evaluate!
  logs(prog_status);
  if(use_cache && precompute_features) es.precompute_features();
  if(use_active_learning)
    es.train_perceptron_with_active_learning();
  else
    es.train_perceptron();

  if(tmp) es.features.dump_debug();

  if(!out_weights_file.empty())
    es.features.print_feature_weights(out_weights_file.c_str(), true);

  return 0;
}
